# -*- coding: utf-8 -*-
"""
Interact with the paltas configuration files

Classes used to draw relevant parameters from paltas configuration files.
"""
import os, sys, warnings, copy
from importlib import import_module
import numpy as np
from ..Sampling.sampler import Sampler
from ..Sources.galaxy_catalog import GalaxyCatalog
from ..Utils.cosmology_utils import get_cosmology, ddt
from ..Utils.hubble_utils import hubblify
from ..Utils.lenstronomy_utils import PSFHelper
from lenstronomy.Data.psf import PSF
from lenstronomy.SimulationAPI.data_api import DataAPI
from lenstronomy.SimulationAPI.observation_api import SingleBand
from lenstronomy.LensModel.lens_model import LensModel
from lenstronomy.LightModel.light_model import LightModel
from lenstronomy.PointSource.point_source import PointSource
from lenstronomy.ImSim.image_model import ImageModel
import lenstronomy.Util.util as util

# Global filters on the python warnings. Using this since filter
# behaviour is a bit weird.
SERIALIZATIONWARNING = True
KWARGSNUMERICWARNING1 = True
KWARGSNUMERICWARNING2 = True


class ConfigHandler():
	"""Class that parses the configuration files to extract images and lenstronomy
	configurations.

	Args:
		config_path (str): A path to the config file to parse.
	"""

	def __init__(self,config_path):
		# Get the dictionary from the provided .py file
		config_dir, config_file = os.path.split(os.path.abspath(config_path))
		sys.path.insert(0, config_dir)
		config_name, _ = os.path.splitext(config_file)
		self.config_module = import_module(config_name)
		self.config_dict = self.config_module.config_dict

		# Initialize the random seed provided with the config module
		if hasattr(self.config_module,'seed'):
			np.random.seed(self.config_module.seed)

		# Set up our sampler and draw a sample for initialization
		self.sampler = Sampler(self.config_dict)
		self.sample = None
		self.draw_new_sample()
		sample = self.get_current_sample()

		# Initialize all the parameters and classes we need for drawing
		# lenstronomy inputs and images

		# Get the numerical kwargs numpix from the config
		self.kwargs_numerics = self.config_module.kwargs_numerics
		self.numpix = self.config_module.numpix

		# Set up the paltas objects we'll use
		self.los_class = None
		self.subhalo_class = None
		self.main_deflector_class = None
		self.lens_light_class = None
		self.point_source_class = None
		self.do_drizzle = False
		if 'los' in self.config_dict:
			self.los_class = self.config_dict['los']['class'](
				sample['los_parameters'],sample['main_deflector_parameters'],
				sample['source_parameters'],sample['cosmology_parameters'])
		if 'subhalo' in self.config_dict:
			self.subhalo_class = self.config_dict['subhalo']['class'](
				sample['subhalo_parameters'],sample['main_deflector_parameters'],
				sample['source_parameters'],sample['cosmology_parameters'])
		if 'main_deflector' in self.config_dict:
			self.main_deflector_class = (
				self.config_dict['main_deflector']['class'](
					sample['main_deflector_parameters'],
					sample['cosmology_parameters']))
		if 'drizzle' in self.config_dict:
			self.do_drizzle = True
		if 'lens_light' in self.config_dict:
			self.lens_light_class = self.config_dict['lens_light']['class'](
				sample['cosmology_parameters'], sample['lens_light_parameters'])
		if 'point_source' in self.config_dict:
			self.point_source_class = self.config_dict['point_source']['class'](
				sample['point_source_parameters'])

		# We always need a source class
		self.source_class = self.config_dict['source']['class'](
			sample['cosmology_parameters'],sample['source_parameters'])

		# See if a magnitude cut was specified
		if hasattr(self.config_module, 'mag_cut'):
			self.mag_cut = self.config_module.mag_cut
		else:
			self.mag_cut = None

		# See if ignoring noise was specified
		if hasattr(self.config_module,'no_noise'):
			self.add_noise = not self.config_module.no_noise
		else:
			self.add_noise = True

	def draw_new_sample(self):
		"""Draws a new sample from the config sampler.
		"""
		self.sample = self.sampler.sample()

	def get_current_sample(self):
		"""Returns the current sample from the config sampler.

		Returns:
			(dict): The current sample
		"""
		return self.sample

	def get_lenstronomy_models_kwargs(self,new_sample=True):
		"""Takes a sample from the config and returns the list of lenstronomy
		models, kwargs, and redshifts for the lensing system.

		Args:
			new_sample (bool): If true will draw a new sample from the config
				sampler before returning the lenstronomy kwargs. True by
				default.

		Returns:
			(dict,dict): Two dicts, the first containing the list of lens
			models, lens model redshifts, source light models, source redshifts,
			lens light models, and point source models. The second contains the
			lens kwargs, the source kwargs, the point source kwargs, and the
			lens light kwargs.

		Notes:
			Even if new_sample is False, this function is not guaranteed to be
			deterministic. For example, most of the substructure classes draw
			from populations specified by the input parameters, and therefore
			calling the function repeatedly will return different realizations
			of that population.
		"""
		# Either draw a new sample or use the current sample.
		if new_sample:
			self.draw_new_sample()
		sample = self.get_current_sample()

		# Populate the list of models and kwargs lenstronomy needs
		complete_lens_model_list = []
		complete_lens_model_kwargs = []
		complete_z_list = []
		lens_light_model_list = []
		lens_light_kwargs_list = []
		point_source_model_list = []
		point_source_kwargs_list = []
		source_model_list = []
		source_kwargs_list = []
		source_redshift_list = []

		# For each lensing object that's present, add them to the model and
		# kwargs list
		if self.los_class is not None:
			self.los_class.update_parameters(
				sample['los_parameters'],sample['main_deflector_parameters'],
				sample['source_parameters'],sample['cosmology_parameters'])
			los_model_list, los_kwargs_list, los_z_list = (
				self.los_class.draw_los())
			interp_model_list, interp_kwargs_list, interp_z_list = (
				self.los_class.calculate_average_alpha(self.numpix*
					self.kwargs_numerics['supersampling_factor']))
			complete_lens_model_list += los_model_list + interp_model_list
			complete_lens_model_kwargs += los_kwargs_list + interp_kwargs_list
			complete_z_list += los_z_list + interp_z_list
		if self.subhalo_class is not None:
			self.subhalo_class.update_parameters(
				sample['subhalo_parameters'],
				sample['main_deflector_parameters'],
				sample['source_parameters'],sample['cosmology_parameters'])
			sub_model_list, sub_kwargs_list, sub_z_list = (
				self.subhalo_class.draw_subhalos())
			complete_lens_model_list += sub_model_list
			complete_lens_model_kwargs += sub_kwargs_list
			complete_z_list += sub_z_list
		if self.main_deflector_class is not None:
			self.main_deflector_class.update_parameters(
				sample['main_deflector_parameters'],sample['cosmology_parameters'])
			main_model_list, main_kwargs_list, main_z_list = (
				self.main_deflector_class.draw_main_deflector())
			complete_lens_model_list += main_model_list
			complete_lens_model_kwargs += main_kwargs_list
			complete_z_list += main_z_list
		if self.lens_light_class is not None:
			self.lens_light_class.update_parameters(
				cosmology_parameters=sample['cosmology_parameters'],
				source_parameters=sample['lens_light_parameters'])
			lens_light_model_list, lens_light_kwargs_list, _ = (
				self.lens_light_class.draw_source())
		if self.point_source_class is not None:
			self.point_source_class.update_parameters(
				sample['point_source_parameters'])
			point_source_model_list,point_source_kwargs_list = (
				self.point_source_class.draw_point_source())

		# Now get the model and kwargs from the source class (which must
		# always be present.)
		self.source_class.update_parameters(
			cosmology_parameters=sample['cosmology_parameters'],
			source_parameters=sample['source_parameters'])

		# For catalog objects we also want to save the catalog index
		# and the (possibly randomized) additional rotation angle. We will
		# therefore push these back into the sample object.
		if isinstance(self.source_class,GalaxyCatalog):
			catalog_i, phi = self.source_class.fill_catalog_i_phi_defaults()
			source_model_list, source_kwargs_list, source_redshift_list = (
				self.source_class.draw_source(catalog_i=catalog_i, phi=phi))
			sample['source_parameters']['catalog_i'] = catalog_i
			sample['source_parameters']['phi'] = phi
		else:
			source_model_list, source_kwargs_list, source_redshift_list = (
				self.source_class.draw_source())

		# Check to see if we need multiplane
		multi_plane = False
		if (len(np.unique(source_redshift_list))>1 or
			len(np.unique(complete_z_list))>1):
			multi_plane = True

		# Package all of the lists into a model and parameters dictionary.
		kwargs_model = {}
		kwargs_params = {}
		kwargs_model['lens_model_list'] = complete_lens_model_list
		kwargs_params['kwargs_lens'] = complete_lens_model_kwargs
		kwargs_model['lens_redshift_list'] = complete_z_list
		kwargs_model['lens_light_model_list'] = lens_light_model_list
		kwargs_params['kwargs_lens_light'] = lens_light_kwargs_list
		kwargs_model['point_source_model_list'] = point_source_model_list
		kwargs_params['kwargs_ps'] = point_source_kwargs_list
		kwargs_model['source_light_model_list'] = source_model_list
		kwargs_params['kwargs_source'] = source_kwargs_list
		kwargs_model['source_redshift_list'] = source_redshift_list
		kwargs_model['multi_plane'] = multi_plane

		# The source convention is definied by the source parameters. This is
		# also what the lens model classes use when setting their parameters.
		kwargs_model['z_source'] = sample['source_parameters']['z_source']
		kwargs_model['z_source_convention'] = kwargs_model['z_source']

		return kwargs_model, kwargs_params

	def get_metadata(self):
		"""Returns the values drawn from the configuration file to generate
		the current sample.

		Returns:
			(dict): A dictionary containing the values drawn from the
			configuration file to generate the current sample. This includes
			the parameters of the lensing system (some of which may be
			population level parameters), the parameters of the observation,
			the cosmology, and any other parameters specified within the
			config_dict of the input configuration file.

		Notes:
			The metadata naming scheme is object_parameters_name_of_parameters.
			For example for the Einstein radius of the main deflector the
			key is main_deflector_parameters_theta_E. For the redshift of
			the source it would be source_parameters_z_source.
		"""
		# Setup the warning filter
		global SERIALIZATIONWARNING

		# Get the samples and the metadata.
		sample = self.get_current_sample()
		metadata = {}

		for component in sample:
			for key in sample[component]:
				comp_value = sample[component][key]
				# Make sure that lists and other objects that cannot be
				# serialized well are not written out. Warn about this only
				# once.
				if isinstance(comp_value,bool):
					metadata[component+'_'+key] = int(comp_value)
				elif (isinstance(comp_value,str) or isinstance(comp_value,int) or
					isinstance(comp_value,float)):
					metadata[component+'_'+key] = comp_value
				elif SERIALIZATIONWARNING:
					warnings.warn('One or more parameters in config_dict '
						'cannot be serialized and will not be written to '
						'metadata.csv',category=RuntimeWarning)
					SERIALIZATIONWARNING = False

		return metadata

	def get_sample_cosmology(self,as_astropy=False):
		"""Return the cosmology object for the current sample.

		Args:
			as_astropy (bool): If True, will return an astropy cosmology
				object instead of a colossus cosmology object. Defaults
				to False.
		Returns:
			(colossus.cosmology.cosmology.Cosmology): An instance of
			the colossus cosmology class. If as_astropy is True, this
			will be an astropy object instead.
		"""
		# Grab the cosmology from the sample
		sample = self.get_current_sample()
		cosmo = get_cosmology(sample['cosmology_parameters'])
		if as_astropy is True:
			return cosmo.toAstropy()
		else:
			return cosmo

	def _calculate_ps_metadata(self,metadata,kwargs_params,point_source_model,
		lens_model):
		"""Calculate time delays and image positions and appends them to
		metadata

		Args:
			metadata (dict): A dictionary containing the metadata that
				will be modified in place.
			lenstronomy_dict (dict):  A dictionary containing the list of model
				kwargs. See get_lenstronomy_model_kwargs.
			point_source_model (lenstronomy.PointSource.point_source.PointSource):
				An instance of the lenstronomy point source model that will be
				used to calculate lensing quantitities.
			lens_model (lenstronomy.LensModel.lens_model.LensModel): An instance
				of the lenstronomy lens model that will be used to calculate
				lensing quantitities.
		"""
		# Extract the sample
		sample = self.get_current_sample()
		cosmo = self.get_sample_cosmology()

		# Calculate image positions
		x_image, y_image = point_source_model.image_position(
			kwargs_params['kwargs_ps'],
			kwargs_params['kwargs_lens'])
		num_images = len(x_image[0])

		# Append to the metadata using the same prefix as the rest of the
		# point source parameters
		pfix = 'point_source_parameters_'
		metadata[pfix+'num_images'] = num_images

		# Calculate magnifications using complete_lens_model
		magnifications = lens_model.magnification(x_image[0],y_image[0],
			kwargs_params['kwargs_lens'])
		# If mag_pert is defined, add that pertubation
		if 'mag_pert' in sample['point_source_parameters'].keys():
			magnifications = magnifications * (
				sample['point_source_parameters']['mag_pert'][
					0:len(magnifications)])

		# Calculate time delays
		if sample['point_source_parameters']['compute_time_delays']:
			if 'kappa_ext' in sample['point_source_parameters'].keys():
				td = lens_model.arrival_time(x_image[0],y_image[0],
					kwargs_params['kwargs_lens'],
					sample['point_source_parameters']['kappa_ext'])
				# Apply errors if defined in config_dict
				if 'time_delay_errors' in (
					sample['point_source_parameters'].keys()):
					errors = sample['point_source_parameters'][
						'time_delay_errors']
					errors = errors[0:len(td)-1]
					td = td + errors
				td -= td[0]
			else:
				raise ValueError('must define kappa_ext in point_source ' +
					'parameters to compute time delays')

			# Calculate time delay distance
			metadata[pfix+'ddt'] = ddt(sample,cosmo)

		# Write out all of the metadata.
		for i in range(0,4):
			if i < num_images:
				metadata[pfix+'x_image_'+str(i)] = x_image[0][i]
				metadata[pfix+'y_image_'+str(i)] = y_image[0][i]
				metadata[pfix+'magnification_'+str(i)] = magnifications[i]
			else:
				metadata[pfix+'x_image_'+str(i)] = np.nan
				metadata[pfix+'y_image_'+str(i)] = np.nan
				metadata[pfix+'magnification_'+str(i)] = np.nan

			if sample['point_source_parameters']['compute_time_delays']:
				if i < len(td):
					metadata[pfix+'time_delay_' + str(i)] = td[i]
				else:
					metadata[pfix+'time_delay_' + str(i)] = np.nan

	def _draw_image_standard(self,add_noise=True,apply_psf=True):
		"""Uses the current config sample to generate an image and the
		associated metadata.

		Args:
			add_noise (bool): If False, noise will not be added. Defaults to
				True.
			apply_psf (bool): If False, the psf will not be applied. Defaults
				to True.
		Returns:
			(np.array,dict): A tuple containing a numpy array of the generated
			image and a metavalue dictionary with the corresponding sampled
			values.
		Notes:
			Will return None,None if the produced image does not meet a cut.
		"""
		# Get the lenstronomy parameters and the sample
		sample = self.get_current_sample()
		kwargs_model, kwargs_params = self.get_lenstronomy_models_kwargs(
			new_sample=False)

		# Get the psf and detector parameters from the sample
		kwargs_psf = sample['psf_parameters']
		kwargs_detector = sample['detector_parameters']

		# Build the psf model
		if apply_psf:
			psf_model = PSF(**kwargs_psf)
		else:
			psf_model = PSF(psf_type='NONE')

		# Build the data and noise models we'll use.
		data_api = DataAPI(numpix=self.numpix,**kwargs_detector)
		single_band = SingleBand(**kwargs_detector)

		# Pull the cosmology and source redshift
		cosmo = get_cosmology(sample['cosmology_parameters'])

		# Build our lens and source models.
		lens_model = LensModel(kwargs_model['lens_model_list'],
			z_source=kwargs_model['z_source'],
			z_source_convention=kwargs_model['z_source_convention'],
			lens_redshift_list=kwargs_model['lens_redshift_list'],
			cosmo=cosmo.toAstropy(),multi_plane=kwargs_model['multi_plane'])
		source_light_model = LightModel(kwargs_model['source_light_model_list'],
			source_redshift_list=kwargs_model['source_redshift_list'])
		lens_light_model = LightModel(kwargs_model['lens_light_model_list'])

		# Point source may need lens eqn solver kwargs
		lens_equation_params = None
		if 'lens_equation_solver_parameters' in sample.keys():
			lens_equation_params = sample['lens_equation_solver_parameters']
		point_source_model = PointSource(
			kwargs_model['point_source_model_list'],lensModel=lens_model,
			save_cache=True,kwargs_lens_eqn_solver=lens_equation_params)

		# Put it together into an image model
		image_model = ImageModel(data_api.data_class,psf_model,
			lens_model,source_light_model,lens_light_model,
			point_source_model,kwargs_numerics=self.kwargs_numerics)

		# Generate our image
		image = image_model.image(kwargs_params['kwargs_lens'],
			kwargs_params['kwargs_source'],
			kwargs_params['kwargs_lens_light'],
			kwargs_params['kwargs_ps'])

		# Check for the magnification cut and apply it.
		if self.mag_cut is not None:
			mag = np.sum(image)/np.sum(source_light_model.total_flux(
				kwargs_params['kwargs_source']))
			if mag < self.mag_cut:
				return None,None

		# If noise is specified, add it.
		if add_noise:
			image += single_band.noise_for_model(image)

		# Extract the metadata from the sample
		metadata = self.get_metadata()

		# If a point source was specified, calculate the time delays
		# and image positions.
		if self.point_source_class is not None:
			self._calculate_ps_metadata(metadata,kwargs_params,
				point_source_model,lens_model)

		return image, metadata

	def _draw_image_drizzle(self):
		"""Uses the current config sample to generate a drizzled image and the
		associated metadata.

		Args:
			apply_psf (bool): If False, the psf will not be applied. Defaults
				to true.
		Returns:
			(np.array,dict): A tuple containing a numpy array of the generated
			image and a metavalue dictionary with the corresponding sampled
			values.
		Notes:
			Will return None,None if the produced image does not meet a cut.
			This function will fail if the drizzle parameters are not
			present.
		"""
		# Grab our warning flags
		global KWARGSNUMERICWARNING1
		global KWARGSNUMERICWARNING2

		# Copy the sample since we will be modifying some parameters
		sample_copy = copy.deepcopy(self.sample)
		kwargs_numerics_copy = copy.deepcopy(self.kwargs_numerics)
		numpix_copy = self.numpix

		# Generate a high resolution version of the image.
		supersample_pixel_scale = self.sample['drizzle_parameters'][
			'supersample_pixel_scale']
		output_pixel_scale = (
			self.sample['drizzle_parameters']['output_pixel_scale'])
		detector_pixel_scale = self.sample['detector_parameters']['pixel_scale']
		offset_pattern = self.sample['drizzle_parameters']['offset_pattern']
		wcs_distortion = self.sample['drizzle_parameters']['wcs_distortion']
		ss_scaling = detector_pixel_scale/supersample_pixel_scale
		self.numpix = int(self.numpix*ss_scaling)

		# Temporairly reset the detector pixel_scale to the supersampled scale.
		self.sample['detector_parameters']['pixel_scale'] = (
			supersample_pixel_scale)

		# Modify the numerics kwargs to account for the supersampled pixel scale
		if 'supersampling_factor' in self.kwargs_numerics:
			if KWARGSNUMERICWARNING1:
				warnings.warn('kwargs_numerics supersampling_factor modified '
					+'for drizzle',category=RuntimeWarning)
				KWARGSNUMERICWARNING1 = False
			self.kwargs_numerics['supersampling_factor'] = max(1,
				int(self.kwargs_numerics['supersampling_factor']/ss_scaling))
		if 'point_source_supersampling_factor' in self.kwargs_numerics:
			if KWARGSNUMERICWARNING2:
				warnings.warn('kwargs_numerics point_source_supersampling_factor'
					+ ' modified for drizzle',category=RuntimeWarning)
				KWARGSNUMERICWARNING2 = False
			self.kwargs_numerics['point_source_supersampling_factor'] = max(1,
				int(self.kwargs_numerics['point_source_supersampling_factor']/
					ss_scaling))

		# Use the normal generation class to make our highres image without
		# noise.
		image_ss, metadata = self._draw_image_standard(add_noise=False,
			apply_psf=False)
		self.sample['detector_parameters']['pixel_scale'] = detector_pixel_scale
		self.numpix = numpix_copy

		# Deal with an image that does not pass a cut.
		if image_ss is None:
			return image_ss, metadata

		# Grab the PSF supersampling factor if present.
		if 'psf_supersample_factor' in self.sample['drizzle_parameters']:
			psf_supersample_factor = (
				self.sample['drizzle_parameters']['psf_supersample_factor'])
		else:
			warnings.warn('No psf_supersample_factor provided so 1 will be ' +
				'assumed.')
			psf_supersample_factor = 1

		if psf_supersample_factor > ss_scaling:
			raise ValueError(f'psf_supersample_factor {psf_supersample_factor} '
				+ f'larger than the supersampling {ss_scaling} defined in the '
				+'drizzle parameters.')

		# Make sure that if the user provided a PIXEL psf that the user specified
		# the same supersampling factor for the drizzle_parameters
		if self.sample['psf_parameters']['psf_type'] == 'PIXEL' and (
			'point_source_supersampling_factor' not in
			self.sample['psf_parameters'] or
			self.sample['psf_parameters']['point_source_supersampling_factor'] !=
			psf_supersample_factor):
			raise ValueError('Must specify point_source_supersampling_factor for '
				'PIXEL psf and the value must equal psf_supersample_factor in the '
				' drizzle parameters')

		# We'll bypass lenstronomy's supersampling code by modifying
		# the data api.
		self.kwargs_numerics['point_source_supersampling_factor'] = 1
		self.kwargs_numerics['supersampling_factor'] = 1
		self.kwargs_numerics['supersampling_convolution'] = False
		self.sample['psf_parameters']['point_source_supersampling_factor'] = 1
		self.sample['detector_parameters']['pixel_scale'] /= (
			psf_supersample_factor)

		# Create our noise and psf models.
		kwargs_detector = self.sample['detector_parameters']
		kwargs_psf = self.sample['psf_parameters']
		single_band = SingleBand(**kwargs_detector)
		data_class = DataAPI(numpix=self.numpix*psf_supersample_factor,
			**kwargs_detector).data_class
		psf_model_lenstronomy = PSF(**kwargs_psf)

		if self.add_noise:
			noise_model = single_band.noise_for_model
		else:
			def noise_model(image):
				return 0

		# Initialize the psf model we'll pass to hubblify.
		psf_model = PSFHelper(data_class,psf_model_lenstronomy,
			self.kwargs_numerics).psf_model

		# Pass the high resolution image to hubblify
		image = hubblify(image_ss,supersample_pixel_scale,detector_pixel_scale,
			output_pixel_scale,noise_model,psf_model,offset_pattern,
			wcs_distortion=wcs_distortion,pixfrac=1.0,kernel='square',
			psf_supersample_factor=psf_supersample_factor)

		# Reset the class properties that were modified.
		self.sample = sample_copy
		self.kwargs_numerics = kwargs_numerics_copy

		# Update the metadata to account for the original sample
		metadata.update(self.get_metadata())

		return image, metadata

	def draw_image(self,new_sample=True):
		"""Takes a sample from the config and generate an image of the strong
		lensing system along with its metadata.

		Args:
			new_sample (bool): If true will draw a new sample from the config
				sampler before returning the image. True by default.

		Returns:
			(np.array,dict): A tuple containing a numpy array of the generated
			image and a metavalue dictionary with the corresponding sampled
			values.

		Notes:
			Even if new_sample is False, this function is not guaranteed to be
			deterministic. For example, most of the substructure classes draw
			from populations specified by the input parameters, and therefore
			calling the function repeatedly will return images of different
			realizations of that population.
		"""
		# Draw a new sample if requested
		if new_sample:
			self.draw_new_sample()

		# Use the appropraite generation function
		if self.do_drizzle:
			image,metadata = self._draw_image_drizzle()
		else:
			# _draw_image_standard has a seperate add_noise parameter so
			# it can be used by _draw_image_drizzle.
			image,metadata = self._draw_image_standard(add_noise=self.add_noise)

		# Mask out an interior region of the image if requested
		if hasattr(self.config_module,'mask_radius') and image is not None:
			kwargs_detector = self.get_current_sample()['detector_parameters']
			x_grid, y_grid = util.make_grid(numPix=image.shape[0],
				deltapix=kwargs_detector['pixel_scale'])
			r = util.array2image(np.sqrt(x_grid**2+y_grid**2))
			image[r<=self.config_module.mask_radius] = 0

		return image,metadata



# -*- coding: utf-8 -*-
"""
Manipulate strong lensing images to be used with tensorflow models.

This module contains the functions that allow the numpy file and csv outputs
of generate to be transformed into a format that can be consumed by
tensorflow.
"""

import numpy as np
import tensorflow as tf
import pandas as pd
import glob, os
from tqdm import tqdm
from lenstronomy.SimulationAPI.observation_api import SingleBand
import warnings
from scipy.ndimage import rotate

# Global filters on the python warnings. Using this since filter
# behaviour is a bit weird.
DEFAULTVALUEWARNING = True


def normalize_outputs(metadata,learning_params,input_norm_path,
	log_learning_params=None):
	""" Normalize the outputs of the network

	Args:
		metadata(pd.DataFrame): A pandas object containing the metadata
		learning_params ([str,...]): A list of strings containing the
			parameters that the network is expected to learn.
		input_norm_path (str): The path to a csv that contains the
			normalization to be applied to the output parameters. If the
			file already exists it will be read, if no file exists it will
			be written.
		log_learning_params ([str,...]): A list of strings containing the
			parameters that the network is expected to learn the log of. Can
			be None.

	Returns:
		(pd.DataFrame): A pandas dataframe with the the mean and standard
		deviation for each parameter.
	"""
	# If normalization file is provided use / write it
	if os.path.isfile(input_norm_path):
		print('Using input normalization found at %s'%(input_norm_path))
		norm_dict = pd.read_csv(input_norm_path,index_col='parameter')
		if not all(elem in norm_dict.index for elem in learning_params):
			raise ValueError('Not all of the learning parameters are ' +
				'present in the input normalization dictionary.')
		if log_learning_params is not None:
			if not all(elem in norm_dict.index for elem in log_learning_params):
				raise ValueError('Not all of the log learning parameters are ' +
					'present in the input normalization dictionary.')
	else:
		print('Writing input normalization to %s'%(input_norm_path))
		norm_dict = pd.DataFrame(columns=['parameter','mean','std'])
		norm_dict['parameter'] = learning_params

		# Calculate the normalization for each parameter
		data = metadata[learning_params].to_numpy()
		norm_dict['mean'] = np.mean(data,axis=0)
		norm_dict['std'] = np.std(data,axis=0)

		# Append a dictionary of the log parameters if provided.
		if log_learning_params is not None:
			log_norm_dict = pd.DataFrame(columns=['parameter','mean','std'])
			log_norm_dict['parameter'] = log_learning_params
			# Calculate the normalization for each parameter
			log_data = metadata[log_learning_params].to_numpy()
			log_norm_dict['mean'] = np.mean(np.log(log_data),axis=0)
			log_norm_dict['std'] = np.std(np.log(log_data),axis=0)
			norm_dict = norm_dict.append(log_norm_dict)

		# Set parameter to the index
		norm_dict = norm_dict.set_index('parameter')
		norm_dict.to_csv(input_norm_path)

	return norm_dict


def unnormalize_outputs(input_norm_path,learning_params,mean,standard_dev=None,
	cov_mat=None):
	""" Given NN outputs, undo the normalization step and return the parameters
	in the original space

	Args:
		input_norm_path (str): The path to a csv that contains the
			normalization to be undone.
		learning_params ([str,...]): A list of strings containing the
			parameters that the network is expected to learn. Length n_params.
		mean (np.array): A numpy array with dimensions (batch_size,n_params)
			containing the mean estimate for each parameter
		standard_dev (np.array): A numpy array with dimensions (batch_size,
			n_params) containing the standard deviation estimate for each
			parameter.
		cov_mat (np.array): A numpy array with dimensions (batch_size,n_params,
			n_params) containing the covariance matrix estiamtes for each
			image.

	Notes:
		All values will be modified in place.
	"""
	# Read our normalization dictionary
	norm_dict = pd.read_csv(input_norm_path,index_col='parameter')
	# Iterate over our parameters
	for lpi, param in enumerate(learning_params):
		param_mean = norm_dict['mean'][param]
		param_std = norm_dict['std'][param]

		# We always want to correct the mean
		mean[:,lpi] *= param_std
		mean[:,lpi] += param_mean

		# If provided we want to correct the standard deviation
		if standard_dev is not None:
			standard_dev[:,lpi] *= param_std

		# If provided we want to correct the covariance matrix
		if cov_mat is not None:
			cov_mat[:,lpi,:] *= param_std
			cov_mat[:,:,lpi] *= param_std


def kwargs_detector_to_tf_noise(image,kwargs_detector):
	""" Add noise to the tf tensor provided in agreement with kwargs_detector

	Args:
		image (tf.Tensor): A tensorflow tensor containing the image
		kwargs_detector (dict): A dictionary containing the detector kwargs
			used to generate the noise on the fly.

	Returns:
		(tf.Tensor): A tensorflow tensor containing the noise.
	"""
	# Use lenstronomy to do our noise calculations (lenstronomy is well
	# tested and is very flexible in its noise inputs).
	single_band = SingleBand(**kwargs_detector)

	# Add the background noise.
	noise = tf.random.normal(tf.shape(image)) * single_band.background_noise

	# Add the poisson noise. We have to redo the functions to be tf friendly
	variance = tf.maximum(single_band.flux_iid(image),tf.zeros_like(image))
	variance = tf.sqrt(variance) / single_band.exposure_time
	if single_band._data_count_unit == 'ADU':
		variance /= single_band.ccd_gain
	noise += tf.random.normal(tf.shape(image)) * variance

	return noise


def generate_tf_record(npy_folder,learning_params,metadata_path,
	tf_record_path):
	""" Generate a TFRecord file from a directory of numpy files.

	Args:
		root_path (str): The path to the folder containing the numpy files.
		learning_params ([str,...]): A list of strings containing the
			parameters that the network is expected to learn.
		metadata_path (str):  The path to the csv file containing the
			image metadata.
		tf_record_path (str): The path to which the TFRecord will be saved
	"""
	# Pull the list of numpy filepaths from the directory
	npy_file_list = glob.glob(os.path.join(npy_folder,'image_*.npy'))
	# Open label csv
	metadata = pd.read_csv(metadata_path, index_col=None)

	# Warn the user a default value of 0 will be used for parameters not
	# present in the metadata file.
	global DEFAULTVALUEWARNING
	for param in learning_params:
		if param not in metadata and DEFAULTVALUEWARNING:
			warnings.warn('One or more parameters in learning_params is not '+
				' present in the metadata. A default value of 0 will be used.',
				category=RuntimeWarning)
			DEFAULTVALUEWARNING = False

	# Initialize the writer object and write the lens data
	with tf.io.TFRecordWriter(tf_record_path) as writer:
		for npy_file in tqdm(npy_file_list):
			# Pull the index from the filename
			index = int(npy_file[-11:-4])
			image_shape = np.load(npy_file).shape
			# The image must be converted to a tf string feature
			image_feature = tf.train.Feature(bytes_list=tf.train.BytesList(
				value=[np.load(npy_file).astype(np.float32).tostring()]))
			# Initialize a feature dictionary with the image, the height,
			# and the width
			feature = {
				'image': image_feature,
				'height': tf.train.Feature(
					int64_list=tf.train.Int64List(value=[image_shape[0]])),
				'width': tf.train.Feature(
					int64_list=tf.train.Int64List(value=[image_shape[1]])),
				'index': tf.train.Feature(
					int64_list=tf.train.Int64List(value=[index]))
			}
			# Add all of the lens parameters to the feature dictionary
			for param in learning_params:
				if param in metadata:
					value = metadata[param][index]
				else:
					value = 0
				# Write the feature
				feature[param] = tf.train.Feature(
					float_list=tf.train.FloatList(value=[value]))
			# Create the tf example object
			example = tf.train.Example(features=tf.train.Features(
				feature=feature))
			# Write out the example to the TFRecord file
			writer.write(example.SerializeToString())


def generate_tf_dataset(tf_record_path,learning_params,batch_size,
	n_epochs,norm_images=False,input_norm_path=None,kwargs_detector=None,
	log_learning_params=None):
	"""	Generate a TFDataset that a model can be trained with.

	Args:
		tf_record_paths (str, or [str,...]) A string specifying the paths to
			the TFRecords that will be used in the dataset. Can also be a list
			of strings for specifying multiple tf_record_paths.
		learning_params ([str,...]): A list of strings containing the
			parameters that the network is expected to learn.
		batch_size (int): The batch size that will be used for training
		n_epochs (int): The number of training epochs. The dataset object will
			deal with iterating over the data for repeated epochs.
		norm_images (bool): If True, images will be normalized to have std 1.
		input_norm_path (str): The path to a csv that contains the
			normalization to be applied to the output parameters. If None
			no normalization will be applied.
		kwargs_detector (dict): A dictionary containing the detector kwargs
			used to generate the noise on the fly. If None no additional
			noise will be added.
		log_learning_params ([str,...]): A list of strings containing the
			parameters that the network is expected to learn the log of. Can
			be None.

	Returns:
		(tf.Dataset): A tf.Dataset object that returns the input image and the
		output labels.

	Notes:
		Do not use kwargs_detector if noise was already added during dataset
		generation.
	"""
	# Read the TFRecords
	raw_dataset = tf.data.TFRecordDataset(tf_record_path)

	# If normalization file is provided use it
	if input_norm_path is not None:
		norm_dict = pd.read_csv(input_norm_path,index_col='parameter')
	else:
		norm_dict = None

	# Load a noise model if requested.
	if kwargs_detector is not None:
		noise_function = kwargs_detector_to_tf_noise
	else:
		warnings.warn('No noise will be added')
		noise_function = None

	# Create the feature decoder that will be used
	def parse_image_features(example):  # pragma: no cover
		data_features = {
			'image': tf.io.FixedLenFeature([],tf.string),
			'height': tf.io.FixedLenFeature([],tf.int64),
			'width': tf.io.FixedLenFeature([],tf.int64),
			'index': tf.io.FixedLenFeature([],tf.int64),
		}
		# Set the log learning params to an empy list if no value is provided.
		if log_learning_params is None:
			log_learning_params_list = []
		else:
			log_learning_params_list = log_learning_params

		for param in learning_params+log_learning_params_list:
			data_features[param] = tf.io.FixedLenFeature([],tf.float32)
		parsed_dataset = tf.io.parse_single_example(example,data_features)
		image = tf.io.decode_raw(parsed_dataset['image'],out_type=float)
		image = tf.reshape(image,(parsed_dataset['height'],
			parsed_dataset['width'],1))

		# Add the noise using the baobab noise function (which is a tf graph)
		if noise_function is not None:
			image += noise_function(image,kwargs_detector)

		# If the images must be normed divide by the std
		if norm_images:
			image = image / tf.math.reduce_std(image)

		# Log the parameter if needed
		for param in log_learning_params_list:
			parsed_dataset[param] = tf.math.log(parsed_dataset[param])

		# Normalize if requested
		if norm_dict is not None:
			for param in learning_params+log_learning_params_list:
				parsed_dataset[param] -= norm_dict['mean'][param]
				parsed_dataset[param] /= norm_dict['std'][param]

		lens_param_values = tf.stack([parsed_dataset[param] for param in
			learning_params+log_learning_params_list])
		return image,lens_param_values

	# Select the buffer size to be slightly larger than the batch
	buffer_size = int(batch_size*1.2)

	# Set the feature decoder as the mapping function. Drop the remainder
	# in the case that batch_size does not divide the number of training
	# points exactly
	dataset = raw_dataset.map(parse_image_features,
		num_parallel_calls=tf.data.experimental.AUTOTUNE).repeat(
		n_epochs).shuffle(buffer_size=buffer_size).batch(batch_size).prefetch(
		tf.data.experimental.AUTOTUNE)
	return dataset


def rotate_params_batch(learning_params,output,rot_angle):
	""" Rotate a batch of lensing parameters according to a specified rotation
	angle.

	Args:
		learning_params ([str,...]): A list of strings containing the
			parameters that the network is expected to learn.
		output (np.array): A numpy array of dimension (batch_size,n_outputs)
			containing the true parameter values for each image in the batch.
			Note that n_outputs should be the same as len(learning_params).
		rot_angle (float): The angle to rotate the image by in radians.

	Notes:
		output is modified in place.
	"""

	# Alter the parameters. Hardcoded for now.
	def rotate_param(x,y,theta):
		return (x*np.cos(theta)-y*np.sin(theta),
			x*np.sin(theta)+y*np.cos(theta))

	if 'main_deflector_parameters_center_x' in learning_params:
		xi = learning_params.index('main_deflector_parameters_center_x')
		yi = learning_params.index('main_deflector_parameters_center_y')
		x,y = rotate_param(output[:,xi],output[:,yi],rot_angle)
		output[:,xi] = x
		output[:,yi] = y
	if 'main_deflector_parameters_e1' in learning_params:
		xi = learning_params.index('main_deflector_parameters_e1')
		yi = learning_params.index('main_deflector_parameters_e2')
		x,y = rotate_param(output[:,xi],output[:,yi],2*rot_angle)
		output[:,xi] = x
		output[:,yi] = y
	if 'main_deflector_parameters_gamma1' in learning_params:
		xi = learning_params.index('main_deflector_parameters_gamma1')
		yi = learning_params.index('main_deflector_parameters_gamma2')
		x,y = rotate_param(output[:,xi],output[:,yi],2*rot_angle)
		output[:,xi] = x
		output[:,yi] = y


def rotate_covariance_batch(learning_params,coavriance_batch,rot_angle):
	""" Rotate a batch of lensing parameters according to a specified rotation
	angle.

	Args:
		learning_params ([str,...]): A list of strings containing the
			parameters that the network is expected to learn.
		coavriance_batch (np.array): A numpy array of dimension
			(batch_size,n_outputs,n_outputs) containing the covariance values
			for each image in the batch.
		rot_angle (float): The angle to rotate the image by in radians.

	Notes:
		coavriance_batch is modified in place.
	"""
	# Calculate rotation matrix. Hardcoded for now.
	def rotation_matrix(theta):
		return np.array([[np.cos(theta),-np.sin(theta)],
			[np.sin(theta),np.cos(theta)]])

	# For each possible pair of parameters, conduct the rotation.
	if 'main_deflector_parameters_center_x' in learning_params:
		xi = learning_params.index('main_deflector_parameters_center_x')
		yi = learning_params.index('main_deflector_parameters_center_y')
		# Ensure the parameters are next to each other. Otherwise this
		# indexing will fail.
		assert xi == yi - 1
		yi+=1
		rot_mat = rotation_matrix(rot_angle)
		coavriance_batch[:,xi:yi,xi:yi] = np.dot(rot_mat,
			np.dot(coavriance_batch[:,xi:yi,xi:yi],rot_mat.T).T).T

	if 'main_deflector_parameters_e1' in learning_params:
		xi = learning_params.index('main_deflector_parameters_e1')
		yi = learning_params.index('main_deflector_parameters_e2')
		# Ensure the parameters are next to each other.
		assert xi == yi - 1
		yi+=1
		rot_mat = rotation_matrix(2*rot_angle)
		coavriance_batch[:,xi:yi,xi:yi] = np.dot(rot_mat,
			np.dot(coavriance_batch[:,xi:yi,xi:yi],rot_mat.T).T).T

	if 'main_deflector_parameters_gamma1' in learning_params:
		xi = learning_params.index('main_deflector_parameters_gamma1')
		yi = learning_params.index('main_deflector_parameters_gamma2')
		assert xi == yi - 1
		yi+=1
		rot_mat = rotation_matrix(2*rot_angle)
		coavriance_batch[:,xi:yi,xi:yi] = np.dot(rot_mat,
			np.dot(coavriance_batch[:,xi:yi,xi:yi],rot_mat.T).T).T


def rotate_image_batch(image_batch,learning_params,output,rot_angle):
	""" Rotate a batch of strong lensing images and the corresponding lensing
	parameters

	Args:
		image_batch (np.array): A numpy image array of shape (batch_size,
			height,width,n_channels) that will be rotated.
		learning_params ([str,...]): A list of strings containing the
			parameters that the network is expected to learn.
		output (np.array): A numpy array of dimension (batch_size,n_outputs)
			containing the true parameter values for each image in the batch.
			Note that n_outputs should be the same as len(learning_params).
		rot_angle (float): The angle to rotate the image by in radians.

	Returns:
		(np.array): A numpy array containing the rotated images.

	Notes:
		output is changed in place.
	"""

	# Rotate the image
	image_batch = rotate(image_batch,-rot_angle*180/np.pi,reshape=False,
		axes=(2,1))

	rotate_params_batch(learning_params,output,rot_angle)

	return image_batch


def generate_rotations_dataset(tf_record_path,learning_params,batch_size,
	n_epochs,norm_images=False,input_norm_path=None,kwargs_detector=None,
	log_learning_params=None):
	"""	Returns a generator that builds off of a TFDataset by adding random
	rotations to the images and parameters.

	Args:
		tf_record_paths (str, or [str,...]) A string specifying the paths to
			the TFRecords that will be used in the dataset. Can also be a list
			of strings for specifying multiple tf_record_paths.
		learning_params ([str,...]): A list of strings containing the
			parameters that the network is expected to learn.
		batch_size (int): The batch size that will be used for training
		n_epochs (int): The number of training epochs. The dataset object will
			deal with iterating over the data for repeated epochs.
		norm_images (bool): If True, images will be normalized to have std 1.
		input_norm_path (str): The path to a csv that contains the
			normalization to be applied to the output parameters. If None
			no normalization will be applied.
		kwargs_detector (dict): A dictionary containing the detector kwargs
			used to generate the noise on the fly. If None no additional
			noise will be added.
		log_learning_params ([str,...]): A list of strings containing the
			parameters that the network is expected to learn the log of. Can
			be None.

	Returns:
		(generator): A generator that returns a tuple with the rotated image
		and the parameter values.
	"""
	# Create our base tf dataset without normalization
	base_dataset = generate_tf_dataset(tf_record_path,learning_params,
		batch_size,n_epochs,norm_images=norm_images,
		kwargs_detector=kwargs_detector,log_learning_params=log_learning_params)

	# If normalization file is provided use it
	if input_norm_path is not None:
		norm_dict = pd.read_csv(input_norm_path,index_col='parameter')
	else:
		norm_dict = None

	# If there are no log_learning_params, set it to an empty list
	if log_learning_params is None:
		log_learning_params = []

	def rotation_generator(dataset):
		# Iterate through the images and parameters in the dataset
		for image_batch, lens_param_batch in dataset:
			image_batch = image_batch.numpy()
			lens_param_batch = lens_param_batch.numpy()
			# Pick a rotation angle
			rot_angle = np.random.uniform()*2*np.pi
			# Conduct the rotation
			image_batch = rotate_image_batch(image_batch,
				learning_params+log_learning_params,lens_param_batch,rot_angle)
			if norm_dict is not None:
				for lpi, param in enumerate(learning_params+log_learning_params):
					lens_param_batch[:,lpi] -= norm_dict['mean'][param]
					lens_param_batch[:,lpi] /= norm_dict['std'][param]
			# Yield the rotated image and parameters
			yield image_batch, lens_param_batch

	# Return a rotation generator on our base dataset.
	return rotation_generator(base_dataset)



# -*- coding: utf-8 -*-
"""
Conduct hierarchical inference on a population of lenses.

This module contains the tools to conduct hierarchical inference on our
network posteriors.
"""
import numpy as np
from scipy import special
import numba


# The predicted samples need to be et as a global variable for the pooling to
# be efficient when done by emcee. This will have shape (num_params,num_samps,
# batch_size).
predict_samps_hier = None

# As with the predicted samples, the predicted mu and cov for the analytical
# calculations should also be set at the global level for optimal
# performance.
mu_pred_array = None
prec_pred_array = None
mu_pred_array_ensemble = None
prec_pred_array_ensemble = None


def log_p_xi_omega(predict_samps_hier,hyperparameters,eval_func_xi_omega):
	""" Calculate log p(xi|omega), the probability of the lens paramaters given
	the proposed lens parameter population level distribution.

	Args:
		predict_samps_hier (np.array): An array with dimensions (num_params,
			num_samps,batch_size) containing samples drawn from the
			predicted distribution.
		hyperparameters (np.array): An array with the proposed hyperparameters
			for the population level lens parameter distribution
		eval_func_xi_omega (function): A callable function with inputs
			(predict_samps_hier,hyperparameters) that returns an array of shape
			(n_samps,batch_size) containing the value of log p(xi|omega) for
			each sample.

	Returns:
		(np.array): A numpy array of the shape (n_samps,batch_size) containing
		the value of log p(xi|omega) for each sample.

	Notes:
		predict_samps_hier has shape (num_params,num_samps,batch_size) to make
		it easier to write fast evaluation functions using numba.
	"""

	# We mostly just need to call eval_func_xi_omega and deal with nans
	logpdf = eval_func_xi_omega(predict_samps_hier,hyperparameters)
	logpdf[np.isnan(logpdf)] = -np.inf

	return logpdf


def log_p_omega(hyperparameters,eval_func_omega):
	""" Calculate log p(omega) - the probability of the hyperparameters given
	the hyperprior.

	Args:
		hyperparameters (np.array): An array with the proposed hyperparameters
			for the population level lens parameter distribution
		eval_func_omega (function): A callable function with inputs
			(hyperparameters) that returns an float equal to the value of
			log p(omega)

	Returns:
		(float): The value of log p(omega)
	"""

	# We mostly need to check for nans.
	logpdf = eval_func_omega(hyperparameters)
	if np.sum(np.isnan(logpdf))>0:
		logpdf = -np.inf

	return logpdf


@numba.njit
def gaussian_product_analytical(mu_pred,prec_pred,mu_omega_i,prec_omega_i,
	mu_omega,prec_omega):  # pragma: no cover
	""" Calculate the log of the integral of p(xi_k|omega)*p(xi_k|d_k,omega_int)/
	p(xi_k|omega_int) when all three pdfs are Gaussian.

	Args:
		mu_pred (np.array): The mean output by the network
		prec_pred (np.array): The precision matrix output by the network
		mu_omega_i (np.array): The mean output of the interim prior
		prec_omega_i (np.array): The precision matrix of the interim prior
		mu_omega (np.array): The mean of the proposed hyperparameter
			posterior.
		prec_omega (np.array): The precision matrix of the proposed
			hyperparameter posterior.

	Returns:
		(float): The lof of the product of the three Gaussian integrated over
		all space.
	Notes:
		The equation used here breaks down when the combination of precision
		matrices does not yield a valid precision matrix. When this happen, the
		output will be -np.inf.
	"""
	# This implements the final formula derived in the appendix of
	# Wagner-Carena et al. 2021.

	# Calculate the values of eta and the combined precision matrix
	prec_comb = prec_pred+prec_omega-prec_omega_i

	# This is not guaranteed to return a valid precision matrix. When it
	# doesn't the analytical equation used here is wrong. In those cases
	# return -np.inf
	eigvals = np.linalg.eigvals(prec_comb.astype(np.complex128))
	if (np.any(np.real(eigvals)<=0) or np.any(np.imag(eigvals)!=0)):
		return -np.inf

	cov_comb = np.linalg.inv(prec_comb)
	eta_pred = np.dot(prec_pred,mu_pred)
	eta_omega_i = np.dot(prec_omega_i,mu_omega_i)
	eta_omega = np.dot(prec_omega,mu_omega)
	eta_comb = eta_pred + eta_omega - eta_omega_i

	# Now calculate each of the terms in our exponent
	exponent = 0
	exponent -= np.log(abs(np.linalg.det(prec_pred)))
	exponent -= np.log(abs(np.linalg.det(prec_omega)))
	exponent += np.log(abs(np.linalg.det(prec_omega_i)))
	exponent += np.log(abs(np.linalg.det(prec_comb)))
	exponent += np.dot(mu_pred.T,np.dot(prec_pred,mu_pred))
	exponent += np.dot(mu_omega.T,np.dot(prec_omega,mu_omega))
	exponent -= np.dot(mu_omega_i.T,np.dot(prec_omega_i,mu_omega_i))
	exponent -= np.dot(eta_comb.T,np.dot(cov_comb,eta_comb))

	return -0.5*exponent


class ProbabilityClass:
	""" A class for the hierarchical inference probability calculations given
	samples from the NN posterior on the data.

	Args:
		eval_func_xi_omega_i (function): A callable function with input
			predict_samps_hier that returns an array of shape (n_samps,n_lenses)
			containing the value of log p(xi|omega_interim) for each sample.
			omega_int is the distribution of the training data.
		eval_func_xi_omega (function): A callable function with inputs
			(predict_samps_hier,hyperparameters) that returns an array of shape
			(n_samps,n_lenses) containing the value of log p(xi|omega) for
			each sample. omega is the proposed distribution of the test data.
		eval_func_omega (function): A callable function with inputs
			(hyperparameters) that returns a float equal to the value of
			log p(omega)
	Notes:
		predict_samps_hier has shape (num_params,num_samps,batch_size) to make
		it easier to write fast evaluation functions using numba.
	"""

	def __init__(self,eval_func_xi_omega_i,eval_func_xi_omega,
		eval_func_omega):
		# Save these functions to the class for later use.
		self.eval_func_xi_omega_i = eval_func_xi_omega_i
		self.eval_func_xi_omega = eval_func_xi_omega
		self.eval_func_omega = eval_func_omega

		self.samples_init = False

	def set_samples(self,predict_samps_input=None,predict_samps_hier_input=None):
		""" Set the global lens samples value. Using a global helps avoid data
		being pickled.

		Args:
			predict_samps_input (np.array): An array of shape (n_samps,n_lenses,
				n_params) that represents the network predictions and will be
				reshaped for hierarchical analysis
			predict_samps_hier_input (np.array): An array of shape (n_params,
				n_samps,n_lenses) that represents the network predictions.

		Notes:
			Both predict_samps_input and predict_samps_hier_input are allowed
			as inputs. The functions in posterior_functions use the predict_samps
			convention (n_samps,n_lenses,n_params) while the functions in this
			package use predict_samps_hier (n_params,n_samps,n_lenses)
			convention.
		"""
		# Set the global samples variable
		global predict_samps_hier
		if predict_samps_hier_input is not None:
			predict_samps_hier = predict_samps_hier_input
		elif predict_samps_input is not None:
			predict_samps_hier = np.ascontiguousarray(np.transpose(
				predict_samps_input,[2,0,1]))
		else:
			raise ValueError('Either predict_samps_input or ' +
				'predict_samps_hier_input must be specified')
		self.samples_init = True

		# Calculate the probability of the sample on the interim training
		# distribution
		self.p_samps_omega_i = self.eval_func_xi_omega_i(predict_samps_hier)

	def log_post_omega(self,hyperparameters):
		""" Given the predicted samples, calculate the log posterior of a
		specific distribution.

		Args:
			hyperparameters (np.array): An array with the proposed
				hyperparameters describing the population level lens parameter
				distribution omega.

		Returns:
			(float): The log posterior of omega given the predicted samples.

		Notes:
			Constant factors are ignored.
		"""

		if self.samples_init is False:
			raise RuntimeError('Must set samples or behaviour is ill-defined.')

		global predict_samps_hier

		# Start with the prior on omega
		lprior = log_p_omega(hyperparameters,self.eval_func_omega)

		# No need to evaluate the samples if the proposal is outside the prior.
		if lprior == -np.inf:
			return lprior

		# Calculate the probability of each datapoint given omega
		p_samps_omega = log_p_xi_omega(predict_samps_hier,hyperparameters,
			self.eval_func_xi_omega)

		# We can use our pre-calculated value of p_samps_omega_i.
		like_ratio = p_samps_omega - self.p_samps_omega_i
		like_ratio[np.isinf(self.p_samps_omega_i)] = -np.inf
		like_ratio = special.logsumexp(like_ratio,axis=0)
		like_ratio[np.isnan(like_ratio)] = -np.inf

		# Return the likelihood and the prior combined
		return lprior + np.sum(like_ratio)


class ProbabilityClassAnalytical:
	""" A class for the hierarchical inference probability calculations that
	works analytically for the case of Gaussian outputs, priors, and target
	distributions.

	Args:
		mu_omega_i (np.array): An array with the length n_params
			specifying the mean of each parameters in the training
			distribution.
		cov_omega_i (np.array): A n_params x n_params array
			specifying the covariance matrix for the training
			distribution.
		eval_func_omega (function): A callable function with inputs
			(hyperparameters) that returns a float equal to the value of
			log p(omega).
	"""
	def __init__(self,mu_omega_i,cov_omega_i,eval_func_omega):
		# Save each parameter to the class
		self.mu_omega_i = mu_omega_i
		self.cov_omega_i = cov_omega_i
		# Store the precision matrix for later use.
		self.prec_omega_i = np.linalg.inv(cov_omega_i)
		self.eval_func_omega = eval_func_omega

		# A flag to make sure the prediction values are set
		self.predictions_init = False

	def set_predictions(self,mu_pred_array_input,prec_pred_array_input):
		""" Set the global lens mean and covariance prediction values.

		Args:
			mu_pred_array_input (np.array): An array of shape (n_lenses,
				n_params) that represents the mean network prediction on each
				lens.
			prec_pred_array_input (np.array): An array of shape (n_lenses,
				n_params,n_params) that represents the predicted precision
				matrix on each lens.
		"""
		# Call up the globals and set them.
		global mu_pred_array
		global prec_pred_array
		mu_pred_array = mu_pred_array_input
		prec_pred_array = prec_pred_array_input

		# Set the flag for the predictions being initialized
		self.predictions_init = True

	@staticmethod
	@numba.njit
	def log_integral_product(mu_pred_array,prec_pred_array,mu_omega_i,
		prec_omega_i,mu_omega,prec_omega):  # pragma: no cover
		""" For the case of Gaussian distributions, calculate the log of the
		integral p(xi_k|omega)*p(xi_k|d_k,omega_int)/p(xi_k|omega_int) summed
		over all of the lenses in the sample.

		Args:
			mu_pred_array (np.array): An array of the mean output by the
				network for each lens
			prec_pred_array (np.array): An array of the precision matrix output
				by the network for each lens.
			prec_pred (np.array): The precision matrix output by the network
			mu_omega_i (np.array): The mean output of the interim prior
			prec_omega_i (np.array): The precision matrix of the interim prior
			mu_omega (np.array): The mean of the proposed hyperparameter
				posterior.
			prec_omega (np.array): The precision matrix of the proposed
				hyperparameter posterior.
		"""
		# In log space, the product over lenses in the posterior becomes a sum
		integral = 0
		for pi in range(len(mu_pred_array)):
			mu_pred = mu_pred_array[pi]
			prec_pred = prec_pred_array[pi]
			integral += gaussian_product_analytical(mu_pred,prec_pred,
				mu_omega_i,prec_omega_i,mu_omega,prec_omega)
		# Treat nan as probability 0.
		if np.isnan(integral):
			integral = -np.inf
		return integral

	def log_post_omega(self,hyperparameters):
		""" Given the predicted means and covariances, calculate the log
		posterior of a specific distribution.

		Args:
			hyperparameters (np.array): An array with the proposed
				hyperparameters describing the population level lens parameter
				distribution omega. Should be length n_params*2, the first
				n_params being the mean and the second n_params being the log
				of the standard deviation for each parameter.

		Returns:
			(float): The log posterior of omega given the predicted samples.

		Notes:
			For now only supports diagonal covariance matrix.
		"""

		if self.predictions_init is False:
			raise RuntimeError('Must set predictions or behaviour is '
				+'ill-defined.')

		global mu_pred_array
		global prec_pred_array

		# Start with the prior on omega
		lprior = log_p_omega(hyperparameters,self.eval_func_omega)

		# No need to evaluate the samples if the proposal is outside the prior.
		if lprior == -np.inf:
			return lprior

		# Extract mu_omega and prec_omega from the provided hyperparameters
		mu_omega = hyperparameters[:len(hyperparameters)//2]
		cov_omega = np.diag(np.exp(hyperparameters[len(hyperparameters)//2:]*2))
		prec_omega = np.linalg.inv(cov_omega)

		like_ratio = self.log_integral_product(mu_pred_array,prec_pred_array,
			self.mu_omega_i,self.prec_omega_i,mu_omega,prec_omega)

		# Return the likelihood and the prior combined
		return lprior + like_ratio


class ProbabilityClassEnsemble(ProbabilityClassAnalytical):
	""" An extension of the class ProbabilityClassAnalytical that allows for
	hierarchical inference with ensemble predictions.

	Args:
		mu_omega_i (np.array): An array with the length n_params
			specifying the mean of each parameters in the training
			distribution.
		cov_omega_i (np.array): A n_params x n_params array
			specifying the covariance matrix for the training
			distribution.
		eval_func_omega (function): A callable function with inputs
			(hyperparameters) that returns a float equal to the value of
			log p(omega).
	"""

	def set_predictions(self,mu_pred_array_input,prec_pred_array_input):
		""" Set the global lens mean and covariance prediction values.

		Args:
			mu_pred_array_input (np.array): An array of shape (n_ensembles,
				n_lenses,n_params) that represents the mean network prediction
				on each lens.
			prec_pred_array_input (np.array): An array of shape (n_ensembles,
				n_lenses,n_params,n_params) that represents the predicted
				precision matrix on each lens.
		"""
		# Call up the globals and set them.
		global mu_pred_array_ensemble
		global prec_pred_array_ensemble
		mu_pred_array_ensemble = mu_pred_array_input
		prec_pred_array_ensemble = prec_pred_array_input

		# Set the flag for the predictions being initialized
		self.predictions_init = True
		self.n_ensemble = len(mu_pred_array_ensemble)

	@staticmethod
	@numba.njit
	def log_integral_product(mu_pred_array,prec_pred_array,mu_omega_i,
		prec_omega_i,mu_omega,prec_omega):  # pragma: no cover
		""" For the case of Gaussian distributions, calculate the log of the
		integral p(xi_k|omega)*p(xi_k|d_k,omega_int)/p(xi_k|omega_int) summed
		over all of the lenses in the sample.

		Args:
			mu_pred_array (np.array): An array of the mean output by the
				network for each lens
			prec_pred_array (np.array): An array of the precision matrix output
				by the network for each lens.
			prec_pred (np.array): The precision matrix output by the network
			mu_omega_i (np.array): The mean output of the interim prior
			prec_omega_i (np.array): The precision matrix of the interim prior
			mu_omega (np.array): The mean of the proposed hyperparameter
				posterior.
			prec_omega (np.array): The precision matrix of the proposed
				hyperparameter posterior.
		"""
		# In log space, the product over lenses in the posterior becomes a sum
		integral = 0
		n_ensemble = len(mu_pred_array)
		for pi in range(mu_pred_array.shape[1]):
			# For each lens, the integral for each ensemble must be summed
			# together.
			ensemble_integral = -np.inf
			for ei in range(mu_pred_array.shape[0]):
				mu_pred = mu_pred_array[ei,pi]
				prec_pred = prec_pred_array[ei,pi]
				ensemble_integral = np.logaddexp(ensemble_integral,
					gaussian_product_analytical(mu_pred,prec_pred,mu_omega_i,
						prec_omega_i,mu_omega,prec_omega))
			# Divide by 1/N_ensemble for each prediction.
			integral += ensemble_integral - np.log(n_ensemble)
		# Treat nan as probability 0.
		if np.isnan(integral):
			integral = -np.inf
		return integral

	def log_post_omega(self,hyperparameters):
		""" Given the predicted means and covariances, calculate the log
		posterior of a specific distribution.

		Args:
			hyperparameters (np.array): An array with the proposed
				hyperparameters describing the population level lens parameter
				distribution omega. Should be length n_params*2, the first
				n_params being the mean and the second n_params being the log
				of the standard deviation for each parameter.

		Returns:
			(float): The log posterior of omega given the predicted samples.

		Notes:
			For now only supports diagonal covariance matrix.
		"""

		if self.predictions_init is False:
			raise RuntimeError('Must set predictions or behaviour is '
				+'ill-defined.')

		global mu_pred_array_ensemble
		global prec_pred_array_ensemble

		# Extract mu_omega and prec_omega from the provided hyperparameters
		mu_omega = hyperparameters[:len(hyperparameters)//2]
		cov_omega = np.diag(np.exp(hyperparameters[len(hyperparameters)//2:]*2))
		prec_omega = np.linalg.inv(cov_omega)

		# Start with the prior on omega
		lprior = log_p_omega(hyperparameters,self.eval_func_omega)

		# No need to evaluate the samples if the proposal is outside the prior.
		if lprior == -np.inf:
			return lprior

		like_ratio = self.log_integral_product(mu_pred_array_ensemble,
			prec_pred_array_ensemble,self.mu_omega_i,self.prec_omega_i,mu_omega,
			prec_omega)

		# Return the likelihood and the prior combined
		return lprior + like_ratio



# -*- coding: utf-8 -*-
"""
Generate simulated strong lensing images using the classes and parameters of
an input configuration dictionary.

This script generates strong lensing images from paltas config dictionaries.

Example
-------
To run this script, pass in the desired config as argument::

	$ python -m generate.py path/to/config.py path/to/save_folder --n 1000

The parameters will be pulled from config.py and the images will be saved in
save_folder. If save_folder doesn't exist it will be created.
"""
import numpy as np
import argparse, os
import shutil
import matplotlib.pyplot as plt
from tqdm import tqdm
import pandas as pd
from paltas.Configs.config_handler import ConfigHandler


def parse_args():
	"""Parse the input arguments by the user

	Returns:
		(argparse.Namespace): An instance of the Namespace object with the
		users provided values.

	"""
	# Initialize the parser and the possible inputs
	parser = argparse.ArgumentParser()
	parser.add_argument('config_dict', help='Path to paltas configuration dict')
	parser.add_argument('save_folder', help='Folder to save images to')
	parser.add_argument('--n', default=1, dest='n', type=int,
		help='Size of dataset to generate (default 1)')
	parser.add_argument('--save_png_too', action='store_true',
		help='Also save a PNG for each image, for debugging')
	parser.add_argument('--tf_record', action='store_true',
		help='Generate the tf record for the training set.')
	args = parser.parse_args()
	return args


def main():
	"""Generates the strong lensing images by drawing parameters values from
	the provided configuration dictionary.
	"""
	# Get the user provided arguments
	args = parse_args()

	# Make the directory if not already there
	if not os.path.exists(args.save_folder):
		os.makedirs(args.save_folder)
	print("Save folder path: {:s}".format(args.save_folder))

	# Copy out config dict
	shutil.copy(
		os.path.abspath(args.config_dict),
		args.save_folder)

	# Use a pandas dataframe to store the parameter values.
	metadata_csv = pd.DataFrame()
	metadata_path = os.path.join(args.save_folder,'metadata.csv')

	# Initialize our config handler
	config_handler = ConfigHandler(args.config_dict)

	# Generate our images
	pbar = tqdm(total=args.n)
	nt = 0
	tries = 0
	while nt < args.n:
		# We always try
		tries += 1

		# Attempt to draw our image
		image, metadata = config_handler.draw_image(new_sample=True)

		# Failed attempt if there is no image output
		if image is None:
			continue

		# Save the image and the metadata
		filename = os.path.join(args.save_folder, 'image_%07d' % nt)
		np.save(filename, image)
		if args.save_png_too:
			plt.imsave(filename + '.png', image)

		metadata_csv = metadata_csv.append(metadata,ignore_index=True)

		# Write out the metadata every 20 images
		if nt == 0:
			# Sort the keys lexographically to ensure consistent writes
			metadata_csv = metadata_csv.reindex(sorted(metadata_csv.columns), axis=1)
			metadata_csv.to_csv(metadata_path, index=None)
			metadata_csv = pd.DataFrame()
		elif nt%20 == 0:
			metadata_csv = metadata_csv.reindex(sorted(metadata_csv.columns), axis=1)
			metadata_csv.to_csv(metadata_path, index=None, mode='a',
				header=None)
			metadata_csv = pd.DataFrame()

		nt += 1
		pbar.update()

	# Make sure anything left in the metadata_csv DataFrame is written out
	metadata_csv = metadata_csv.reindex(sorted(metadata_csv.columns), axis=1)
	metadata_csv.to_csv(metadata_path, index=None, mode='a',header=None)
	pbar.close()
	print('Dataset generation complete. Acceptance rate: %.3f'%(args.n/tries))

	# Generate tf record if requested. Save all the parameters and use default
	# filename data.tfrecord
	if args.tf_record:
		# Delayed import, triggers tensorflow import
		from paltas.Analysis import dataset_generation

		# The path to save the TFRecord to.
		tf_record_path = os.path.join(args.save_folder,'data.tfrecord')
		# Generate the list of learning parameters. Only save learning
		# parameters with associated float values.
		learning_params = []
		for key in metadata:
			if (isinstance(metadata[key],float) or
				isinstance(metadata[key],int)):
				learning_params.append(key)
		# Generate the TFRecord
		dataset_generation.generate_tf_record(args.save_folder,learning_params,
			metadata_path,tf_record_path)


if __name__ == '__main__':
	main()



# Configuration for a tight spread on main deflector with PEMD + SHEAR, sources
# drawn from COSMOS, and only varying the d_los and sigma_sub of the DG_19
# subhalo and los classes

import numpy as np
from scipy.stats import norm, truncnorm, uniform
from paltas.Substructure.los_dg19 import LOSDG19
from paltas.Substructure.subhalos_dg19 import SubhalosDG19
from paltas.MainDeflector.simple_deflectors import PEMDShear
from paltas.Sources.cosmos import COSMOSExcludeCatalog
from lenstronomy.Util.kernel_util import degrade_kernel
from paltas.Sampling import distributions
from astropy.io import fits
import pandas as pd
import paltas
import os

# Define the numerics kwargs.
kwargs_numerics = {'supersampling_factor':2,'supersampling_convolution':True}
# We do not use point_source_supersampling_factor but it must be passed in to
# surpress a warning.
kwargs_numerics['point_source_supersampling_factor'] = (
	kwargs_numerics['supersampling_factor'])
# This is always the number of pixels for the CCD. If drizzle is used, the
# final image will be larger.
numpix = 128

# Define some general image kwargs for the dataset
mask_radius = 0.5
mag_cut = 2.0

# Define arguments that will be used multiple times
output_ab_zeropoint = 25.127

# Define the cosmos path
root_path = paltas.__path__[0][:-7]
cosmos_folder = root_path + r'/datasets/cosmos/COSMOS_23.5_training_sample/'

# Load the empirical psf. Grab a psf from the middle of the first chip.
# Degrade to account for the 4x supersample
hdul = fits.open(os.path.join(root_path,
	'datasets/hst_psf/emp_psf_f814w.fits'))
# Don't leave any 0 values in the psf.
psf_pix_map = degrade_kernel(hdul[0].data[17]-np.min(hdul[0].data[17]),2)

config_dict = {
	'subhalo':{
		'class': SubhalosDG19,
		'parameters':{
			'sigma_sub':norm(loc=2e-3,scale=1.1e-3).rvs,
			'shmf_plaw_index':uniform(loc=-1.92,scale=0.1).rvs,
			'm_pivot': 1e10,'m_min': 1e7,'m_max': 1e10,
			'c_0':uniform(loc=16,scale=2).rvs,
			'conc_zeta':uniform(loc=-0.3,scale=0.1).rvs,
			'conc_beta':uniform(loc=0.55,scale=0.3).rvs,
			'conc_m_ref': 1e8,
			'dex_scatter': uniform(loc=0.1,scale=0.06).rvs,
			'k1':0.0, 'k2':0.0
		}
	},
	'los':{
		'class': LOSDG19,
		'parameters':{
			'delta_los':norm(loc=1,scale=0.6).rvs,
			'm_min':1e7,'m_max':1e10,'z_min':0.01,
			'dz':0.01,'cone_angle':8.0,'r_min':0.5,'r_max':10.0,
			'c_0':uniform(loc=16,scale=2).rvs,
			'conc_zeta':uniform(loc=-0.3,scale=0.1).rvs,
			'conc_beta':uniform(loc=0.55,scale=0.3).rvs,
			'conc_m_ref': 1e8,
			'dex_scatter': uniform(loc=0.1,scale=0.06).rvs,
			'alpha_dz_factor':5.0
		}
	},
	'main_deflector':{
		'class': PEMDShear,
		'parameters':{
			'M200': 1e13,
			'z_lens': 0.5,
			'gamma': truncnorm(-20,np.inf,loc=2.0,scale=0.1).rvs,
			'theta_E': truncnorm(-1.1/0.15,np.inf,loc=1.1,scale=0.15).rvs,
			'e1': norm(loc=0.0,scale=0.1).rvs,
			'e2': norm(loc=0.0,scale=0.1).rvs,
			'center_x': norm(loc=0.0,scale=0.16).rvs,
			'center_y': norm(loc=0.0,scale=0.16).rvs,
			'gamma1': norm(loc=0.0,scale=0.05).rvs,
			'gamma2': norm(loc=0.0,scale=0.05).rvs,
			'ra_0':0.0, 'dec_0':0.0
		}
	},
	'source':{
		'class': COSMOSExcludeCatalog,
		'parameters':{
			'z_source':1.5,'cosmos_folder':cosmos_folder,
			'max_z':1.0,'minimum_size_in_pixels':64,'faintest_apparent_mag':20,
			'smoothing_sigma':0.00,'random_rotation':True,
			'output_ab_zeropoint':output_ab_zeropoint,
			'center_x':norm(loc=0.0,scale=0.16).rvs,
			'center_y':norm(loc=0.0,scale=0.16).rvs,
			'min_flux_radius':10.0,'source_exclusion_list':np.append(
				pd.read_csv(
					os.path.join(root_path,'paltas/Sources/bad_galaxies.csv'),
					names=['catalog_i'])['catalog_i'].to_numpy(),
				pd.read_csv(
					os.path.join(root_path,'paltas/Sources/val_galaxies.csv'),
					names=['catalog_i'])['catalog_i'].to_numpy())}
	},
	'cosmology':{
		'parameters':{
			'cosmology_name': 'planck18'
		}
	},
	'psf':{
		'parameters':{
			'psf_type':'PIXEL',
			'kernel_point_source': psf_pix_map,
			'point_source_supersampling_factor':2
		}
	},
	'detector':{
		'parameters':{
			'pixel_scale':0.040,'ccd_gain':1.58,'read_noise':3.0,
			'magnitude_zero_point':output_ab_zeropoint,
			'exposure_time':1380,'sky_brightness':21.83,
			'num_exposures':1,'background_noise':None
		}
	},
	'drizzle':{
		'parameters':{
			'supersample_pixel_scale':0.020,'output_pixel_scale':0.030,
			'wcs_distortion':None,
			'offset_pattern':[(0,0),(0.5,0),(0.0,0.5),(-0.5,-0.5)],
			'psf_supersample_factor':2
		}
	},
	'cross_object':{
		'parameters':{
			'subhalo:c_0,los:c_0':distributions.Duplicate(
				dist=uniform(loc=16,scale=2).rvs),
			'subhalo:conc_zeta,los:conc_zeta':distributions.Duplicate(
				dist=uniform(loc=-0.3,scale=0.1).rvs),
			'subhalo:conc_beta,los:conc_beta':distributions.Duplicate(
				dist=uniform(loc=0.55,scale=0.3).rvs),
			'subhalo:dex_scatter,los:dex_scatter':distributions.Duplicate(
				dist=uniform(loc=0.1,scale=0.06).rvs)
		}
	}
}



# -*- coding: utf-8 -*-
"""
Draw subhalo masses and concentrations for NFW subhalos according to
https://arxiv.org/pdf/1909.02573.pdf

This module contains the functions needed to turn the parameters of NFW
subhalo distributions into masses, concentrations, and positions for those
NFW subhalos.
"""
from .subhalos_base import SubhalosBase
from . import nfw_functions
import numba
import numpy as np
from ..Utils import power_law, cosmology_utils
from colossus.halo.concentration import peaks


class SubhalosDG19(SubhalosBase):
	"""Class for rendering the subhalos of a main halos according to DG19.

	Args:
		subhalo_parameters (dict): A dictionary containing the type of
			subhalo distribution and the value for each of its parameters.
		main_deflector_parameters (dict): A dictionary containing the type of
			main deflector and the value for each of its parameters.
		source_parameters (dict): A dictionary containing the type of the
			source and the value for each of its parameters.
		cosmology_parameters (str,dict, or colossus.cosmology.Cosmology):
			Either a name of colossus cosmology, a dict with 'cosmology name':
			name of colossus cosmology, an instance of colussus cosmology, or a
			dict with H0 and Om0 ( other parameters will be set to defaults).

	Notes:

	Required Parameters

	- sigma_sub - SHMF normalization in units of kpc^(-2)
	- shmf_plaw_index - SHMF mass function power-law slope
	- m_pivot - SHMF power-law pivot mass in unit of M_sun
	- m_min - SHMF minimum rendered mass in units of M_sun
	- m_max - SHMF maximum rendered mass in units of M_sun
	- c_0 - concentration normalization
	- conc_zeta - concentration redshift power law slope
	- conc_beta - concentration peak height power law slope
	- conc_m_ref - concentration peak height pivot mass
	- dex_scatter - scatter in concentration in units of dex
	- k1 - slope of SHMF host mass dependence
	- k2 - slope of SHMF host redshift dependence
	"""
	# Define the parameters we expect to find for the DG_19 model
	required_parameters = ('sigma_sub','shmf_plaw_index','m_pivot','m_min',
		'm_max','c_0','conc_zeta','conc_beta','conc_m_ref','dex_scatter',
		'k1','k2')

	def __init__(self,subhalo_parameters,main_deflector_parameters,
		source_parameters,cosmology_parameters):

		# Initialize the super class
		super().__init__(subhalo_parameters,main_deflector_parameters,
			source_parameters,cosmology_parameters)

	@staticmethod
	@numba.njit()
	def host_scaling_function(host_m200, z_lens, k1=0.88, k2=1.7):
		"""Returns scaling for the subhalo mass function based on the mass of
		the host halo.

		Derived from galacticus in https://arxiv.org/pdf/1909.02573.pdf.

		Args:
			host_m200 (float): The mass of the host halo in units of M_sun
			z_lens (flat): The redshift of the host halo / main deflector
			k1 (flaot): Amplitude of halo mass dependence
			k2 (flaot): Amplitude of the redshift scaling

		Returns:
			(float): The normalization scaling for the subhalo mass function

		Notes:
			Default values of k1 and k2 are derived from galacticus.
		"""
		# Equation from DG_19
		log_f = k1 * np.log10(host_m200/1e13) + k2 * np.log10(z_lens+0.5)
		return 10**log_f

	def draw_nfw_masses(self):
		"""Draws from the https://arxiv.org/pdf/1909.02573.pdf subhalo mass
		function and returns an array of the masses.

		Returns:
			(np.array): The masses of the drawn halos in units of M_sun
		"""

		# Pull the parameters we need from the input dictionaries
		# Units of m_sun times inverse kpc^2
		sigma_sub = max(0, self.subhalo_parameters['sigma_sub'])
		shmf_plaw_index = self.subhalo_parameters['shmf_plaw_index']
		# Units of m_sun
		m_pivot = self.subhalo_parameters['m_pivot']
		# Units of m_sun
		host_m200 = self.main_deflector_parameters['M200']
		# Units of m_sun
		m_min = self.subhalo_parameters['m_min']
		# Units of m_sun
		m_max = self.subhalo_parameters['m_max']
		z_lens = self.main_deflector_parameters['z_lens']
		k1 = self.subhalo_parameters['k1']
		k2 = self.subhalo_parameters['k2']

		# Calculate the overall norm of the power law. This includes host
		# scaling, sigma_sub, and the area of interest.
		f_host = self.host_scaling_function(host_m200,z_lens,k1=k1,k2=k2)

		# In DG_19 subhalos are rendered up until 3*theta_E.
		# Colossus return in MPC per h per radian so must be converted to kpc
		# per arc second
		kpc_per_arcsecond = cosmology_utils.kpc_per_arcsecond(z_lens,
			self.cosmo)
		r_E = (kpc_per_arcsecond*self.main_deflector_parameters['theta_E'])
		dA = np.pi * (3*r_E)**2

		# We can also fold in the pivot mass into the norm for simplicity (then
		# all we have to do is sample from a power law).
		norm = f_host*dA*sigma_sub*m_pivot**(-shmf_plaw_index-1)

		# Draw from our power law and return the masses.
		masses = power_law.power_law_draw(m_min,m_max,shmf_plaw_index,norm)
		return masses

	def mass_concentration(self,z,m_200):
		"""Returns the concentration of halos at a certain mass given the
		parameterization of DG_19.

		Args:
			z (np.array): The redshift of the nfw halo
			m_200 (np.array): array of M_200 of the nfw halo units of M_sun

		Returns:
			(np.array): The concentration for each halo.
		"""
		# Get the concentration parameters
		c_0 = self.subhalo_parameters['c_0']
		zeta = self.subhalo_parameters['conc_zeta']
		beta = self.subhalo_parameters['conc_beta']
		m_ref = self.subhalo_parameters['conc_m_ref']
		dex_scatter = self.subhalo_parameters['dex_scatter']

		# The peak calculation is done by colossus. The cosmology must have
		# already been set. Note these functions expect M_sun/h units (which
		# you get by multiplying by h
		# https://www.astro.ljmu.ac.uk/~ikb/research/h-units.html)
		h = self.cosmo.h
		peak_heights = peaks.peakHeight(m_200*h,z)
		peak_height_ref = peaks.peakHeight(m_ref*h,0)

		# Now get the concentrations and add scatter
		concentrations = c_0*(1+z)**(zeta)*(peak_heights/peak_height_ref)**(
			-beta)
		if isinstance(concentrations,np.ndarray):
			conc_scatter = np.random.randn(len(concentrations))*dex_scatter
		elif isinstance(concentrations,float):
			conc_scatter = np.random.randn()*dex_scatter
		concentrations = 10**(np.log10(concentrations)+conc_scatter)

		return concentrations

	@staticmethod
	def rejection_sampling(r_samps,r_200,r_3E):
		"""Given the radial sampling of the positions and DG_19 constraints,
		conducts rejection sampling and return the cartesian positions.

		Args:
			r_samps (np.array): Samples of the radial coordinates for
				the subhalos in units of kpc.
			r_200 (float): The r_200 of the host halo which will be used
				as the maximum z magnitude in units of kpc.
			r_3E (float): 3 times the einstein radius, which will be used
				to bound the x and y coordinates in units of kpc.

		Returns:
			([np.array,...]): A list of two numpy arrays: the boolean
			array of accepted samples and a n_subsx3 array of x,y,z
			coordinates. All in units of kpc.
		"""
		# Sample theta and phi values for all of the radii samples
		theta = np.random.rand(len(r_samps)) * 2 * np.pi
		phi = np.arccos(1-2*np.random.rand(len(r_samps)))

		# Initialize the x,y,z array
		cart_pos = np.zeros(r_samps.shape+(3,))

		# Get the x, y, and z coordinates
		cart_pos[:,0] += r_samps*np.sin(phi)*np.cos(theta)
		cart_pos[:,1] += r_samps*np.sin(phi)*np.sin(theta)
		cart_pos[:,2] += r_samps*np.cos(phi)

		# Test which samples are outside the DG_19 bounds
		r2_inside = np.sqrt(cart_pos[:,0]**2+cart_pos[:,1]**2)<r_3E
		z_inside = np.abs(cart_pos[:,2])<r_200
		keep = np.logical_and(r2_inside,z_inside)

		return (keep,cart_pos)

	def sample_cored_nfw(self,n_subs):
		"""Given the a tidal radius that defines a core region and the
		parameters of the main deflector, samples positions for NFW subhalos
		bounded as described in https://arxiv.org/pdf/1909.02573.pdf

		Args:
			n_subs (int): The number of subhalo positions to sample

		Returns:
			(np.array): A n_subs x 3 array giving the x,y,z position of the
			subhalos in units of kpc.

		Notes:
			The code works through rejection sampling, which can be inneficient
			for certain configurations. If this is a major issue, it may be
			worth introducing more analytical components.
		"""

		# Create an array that will store our coordinates
		cart_pos = np.zeros((n_subs,3))

		# Calculate the needed host properties
		host_m200 = self.main_deflector_parameters['M200']
		z_lens = self.main_deflector_parameters['z_lens']
		host_c = self.mass_concentration(z_lens,host_m200)
		host_r_200 = nfw_functions.r_200_from_m(host_m200,z_lens,self.cosmo)
		host_r_scale = host_r_200/host_c
		# DG_19 definition of the tidal radius
		r_tidal = host_r_200/2
		host_rho_nfw = nfw_functions.rho_nfw_from_m_c(host_m200,host_c,
			self.cosmo,r_scale=host_r_scale)

		# Tranform the einstein radius to physical units (TODO this should
		# be a function). Multiply by 3 since that's what's relevant for
		# DG_19 parameterization.
		kpc_per_arcsecond = cosmology_utils.kpc_per_arcsecond(z_lens,self.cosmo)
		r_3E = (kpc_per_arcsecond*self.main_deflector_parameters['theta_E'])*3

		# The largest radius we should bother sampling is set by the diagonal of
		# our cylinder.
		r_max = np.sqrt(r_3E**2+host_r_200**2)

		n_accepted_draws = 0
		r_subs = nfw_functions.cored_nfw_draws(r_tidal,host_rho_nfw,
			host_r_scale,r_max,n_subs)
		keep_ind, cart_draws = self.rejection_sampling(r_subs,host_r_200,r_3E)

		# Save the cartesian coordinates we want to keep
		cart_pos[n_accepted_draws:n_accepted_draws+np.sum(keep_ind)] = (
			cart_draws[keep_ind])
		n_accepted_draws += np.sum(keep_ind)

		# Get the fraction of rejection to see how much we should sample
		rejection_frac = max(1-np.mean(keep_ind),1e-1)

		# Keep drawing until we have enough r_subs.
		while n_accepted_draws<n_subs:
			r_subs = nfw_functions.cored_nfw_draws(r_tidal,host_rho_nfw,
				host_r_scale,r_max,int(np.round(n_subs*rejection_frac)))
			keep_ind, cart_draws = self.rejection_sampling(r_subs,host_r_200,
				r_3E)
			use_keep = np.minimum(n_subs-n_accepted_draws,np.sum(keep_ind))
			# Save the cartesian coordinates we want to keep
			cart_pos[n_accepted_draws:n_accepted_draws+use_keep] = (
				cart_draws[keep_ind][:use_keep])
			n_accepted_draws += use_keep

		return cart_pos

	@staticmethod
	def get_truncation_radius(m_200,r,m_pivot=1e7,r_pivot=50):
		"""Returns the truncation radius for a subhalo given the mass and
		radial position in the host NFW

		Args:
			m_200 (np.array): The mass of the subhalos in units of M_sun
			r (np.array): The radial position of the subhalos in units of kpc
			m_pivot (float): The pivot mass for the scaling in units of M_sun
			r_pivot (float): The pivot radius for the relation in unit of kpc

		Returns:
			(np.array): The truncation radii for the subhalos in units of kpc
		"""

		return 1.4*(m_200/m_pivot)**(1/3)*(r/r_pivot)**(2/3)

	def convert_to_lenstronomy(self,subhalo_masses,subhalo_cart_pos):
		"""Converts the subhalo masses and position to truncated NFW profiles
		for lenstronomy

		Args:
			subhalo_masses (np.array): The masses of each of the subhalos that
				were drawn
			subhalo_cart_pos (np.array): A n_subs x 3D array of the positions
				of the subhalos that were drawn
		Returns:
			([string,...],[dict,...]): A tuple containing the list of models
			and the list of kwargs for the truncated NFWs.
		"""
		# First, for each subhalo mass we'll also have to draw a concentration.
		# This requires a redshift. DG_19 used the predicted redshift of infall
		# from galacticus. For now, we'll use the redshift of the lens itself.
		# TODO: Use a different redshift
		z_lens = self.main_deflector_parameters['z_lens']
		z_source = self.source_parameters['z_source']
		subhalo_z = (np.ones(subhalo_masses.shape) *
			self.main_deflector_parameters['z_lens'])
		concentration = self.mass_concentration(subhalo_z,subhalo_masses)

		# We'll also need the radial position in the halo
		r_in_host = np.sqrt(np.sum(subhalo_cart_pos**2,axis=-1))

		# Now we can convert these masses and concentrations into NFW parameters
		# for lenstronomy.
		sub_r_200 = nfw_functions.r_200_from_m(subhalo_masses,subhalo_z,
			self.cosmo)
		sub_r_scale = sub_r_200/concentration
		sub_rho_nfw = nfw_functions.rho_nfw_from_m_c(subhalo_masses,
			concentration,self.cosmo,
			r_scale=sub_r_scale)
		sub_r_trunc = self.get_truncation_radius(subhalo_masses,r_in_host)

		# Convert to lenstronomy units
		sub_r_scale_ang, alpha_Rs, sub_r_trunc_ang = (
			nfw_functions.convert_to_lenstronomy_tNFW(
				sub_r_scale,subhalo_z,sub_rho_nfw,sub_r_trunc,z_source,
				self.cosmo))
		kpc_per_arcsecond = cosmology_utils.kpc_per_arcsecond(z_lens,
			self.cosmo)
		cart_pos_ang = subhalo_cart_pos / np.expand_dims(kpc_per_arcsecond,
			axis=-1)

		# Populate the parameters for each lens
		model_list = []
		kwargs_list = []
		for i in range(len(subhalo_masses)):
			model_list.append('TNFW')
			kwargs_list.append({'alpha_Rs':alpha_Rs[i],'Rs':sub_r_scale_ang[i],
				'center_x':(cart_pos_ang[i,0]+
					self.main_deflector_parameters['center_x']),
				'center_y':(cart_pos_ang[i,1]+
					self.main_deflector_parameters['center_y']),
				'r_trunc':sub_r_trunc_ang[i]})

		return (model_list,kwargs_list)

	def draw_subhalos(self):
		"""Draws masses, concentrations,and positions for the subhalos of a
		main lens halo.

		Returns:
			(tuple): A tuple of three lists: the first is the profile type for
			each subhalo returned, the second is the lenstronomy kwargs for
			that subhalo, and the third is the redshift for each subhalo.
		Notes:
			The redshift for each subhalo is the same as the host, so the
			returned redshift list is not necessary unless the output is
			being combined with los substructure.
		"""
		# Initialize the lists that will contain our mass profile types and
		# assosciated kwargs. If no subhalos are drawn, these will remain empty
		subhalo_model_list = []
		subhalo_kwargs_list = []

		# Distribute subhalos according to https://arxiv.org/pdf/1909.02573.pdf
		# DG_19 assumes NFWs distributed throughout the main deflector.
		# For these NFWs we need positions, masses, and concentrations that
		# we will then translate to Lenstronomy parameters.
		subhalo_masses = self.draw_nfw_masses()

		# It is possible for there to be no subhalos. In that regime
		# just return empty lists
		if subhalo_masses.size == 0:
			return (subhalo_model_list, subhalo_kwargs_list, [])

		subhalo_cart_pos = self.sample_cored_nfw(len(subhalo_masses))
		model_list, kwargs_list = self.convert_to_lenstronomy(
			subhalo_masses,subhalo_cart_pos)
		subhalo_model_list += model_list
		subhalo_kwargs_list += kwargs_list
		subhalo_z_list = [self.main_deflector_parameters['z_lens']]*len(
			subhalo_masses)

		return (subhalo_model_list, subhalo_kwargs_list, subhalo_z_list)



# -*- coding: utf-8 -*-
"""
Define the class to draw line of sight substructure for a lens according to
https://arxiv.org/pdf/1909.02573.pdf

This module contains the functions needed to turn the parameters of the
los halo distribution into masses, concentrations, and positions.
"""
from .los_base import LOSBase
import numba
import numpy as np
from colossus.lss import peaks, bias
from ..Utils import power_law, cosmology_utils
import functools
from . import nfw_functions
import lenstronomy.Util.util as util
from lenstronomy.LensModel.Profiles.nfw import NFW
from scipy.signal import fftconvolve


class LOSDG19(LOSBase):
	"""Class for rendering the line of sight structure according to DG19.

	Args:
		los_parameters (dict): A dictionary containing the type of
			los distribution and the value for each of its parameters.
		main_deflector_parameters (dict): A dictionary containing the type of
			main deflector and the value for each of its parameters.
		source_parameters (dict): A dictionary containing the type of the
			source and the value for each of its parameters.
		cosmology_parameters (str,dict, or colossus.cosmology.Cosmology):
			Either a name of colossus cosmology, a dict with 'cosmology name':
			name of colossus cosmology, an instance of colussus cosmology, or a
			dict with H0 and Om0 ( other parameters will be set to defaults).

	Notes:

	Required Parameters

	- delta_los - mass function normalization
	- m_min - minimum rendered mass in units of M_sun
	- m_max - maximum rendered mass in units of M_sun
	- z_min - minimum redshift at which to render los halos
	- dz - redshift bin width
	- cone_angle - cone opening angle for rendering los halos
	- c_0 - concentration normalization
	- conc_zeta - concentration redshift power law slope
	- conc_beta - concentration peak height power law slope
	- conc_m_ref - concentration peak height pivot mass
	- dex_scatter - scatter in concentration in units of dex
	- alpha_dz_factor- deflection angle correction redshift bin width
	"""
	# Define the parameters we expect to find for the DG_19 model
	required_parameters = ('m_min','m_max','z_min','dz','cone_angle',
		'r_max','r_min','c_0','conc_zeta','conc_beta','conc_m_ref',
		'dex_scatter','delta_los','alpha_dz_factor')

	def __init__(self,los_parameters,main_deflector_parameters,
		source_parameters,cosmology_parameters):

		# Initialize the super class
		super().__init__(los_parameters,main_deflector_parameters,
			source_parameters,cosmology_parameters)

	@staticmethod
	@numba.njit
	def nu_f_nu(nu):
		"""Calculates nu f(nu) for the Sheth Tormen 2001 model.

		Args:
			nu (np.array): An array of nu values to at which to
				calculate nu_f_nu

		Returns:
			(np.array): The value of nu f(nu)

		"""
		# Parameters fit to simulations (A is fixed so that the integral
		# for all nu returns 1).
		A = 0.32218
		q = 0.3
		a = 0.707
		# Calculate the fundamental unit of our equation
		nu_2 = a*np.square(nu)
		# Finally calculate and return nu f(nu)
		nu_f_nu = (2*A*(1+nu_2**(-q))*np.sqrt(nu_2/(2*np.pi))*np.exp(-nu_2/2))
		return nu_f_nu

	def dn_dm(self,m,z):
		"""Returns the mass function at a given mass and redshift.

		Args:
			m (np.array): An array of the mass values at which to calculate
				the mass function in units of M_sun.
			z (np.array): Either one redshift at which to calculate the
				mass function or one redshift for each mass input.

		Returns:
			(np.array): The values of the mass function at each mass in units
			of physical number density in units of 1/(M_sun*kpc^3).

		Notes:
			It is slower to pass in multiple z values that are identical.
			If only a few redshift bins are being considered, run this
			function on each bin with a float redshift.
		"""
		# First we need to do some unit conversaion to the units used by the
		# nu_f_nu function. Note that these are all comoving values we're
		# using because colossus expects them, but nu is dimensionless.
		h = self.cosmo.h
		delta_c = peaks.collapseOverdensity(z=z,corrections=True)
		r = peaks.lagrangianR(m*h)
		sigma = self.cosmo.sigma(r,z)
		nu = delta_c/sigma

		# Calcualte our mass function from its parts
		nfn_eval = self.nu_f_nu(nu)
		d_ln_sigma_d_ln_r = self.cosmo.sigma(r,z,derivative=True)
		# Density is returned in units of M_sun*h^2/kpc^3
		rho_m = self.cosmo.rho_m(z)*h**2

		return -1/3*nfn_eval*rho_m/m**2*d_ln_sigma_d_ln_r

	@functools.lru_cache(maxsize=None)
	def power_law_dn_dm(self,z,m_min,m_max,n_dm=100):
		"""Returns the best fit power law parameters for the physical number
		density at a given redshift and mass range.

		Args:
			z (float): The redshift at which to calculate the power law
				parameters.
			m_min (float): The lower bound of the mass M_sun
			m_max (float): The upper bound of the mass M_sun
			n_dm (int): The number of dm samples to consider in the fit.

		Returns:
			(tuple): The power law slope and the norm for the power law in
			units of 1/M_sun/kpc**3
		"""
		m = np.logspace(np.log10(m_min),np.log10(m_max),n_dm)
		lm = np.log(m)
		dn_dm = self.dn_dm(m,z)
		ldn_dm = np.log(dn_dm)

		# The MLE estimate from the slope assuming Gaussian noise on the
		# log quantity
		slope_estimate = 1/n_dm * np.sum(ldn_dm) * np.sum(lm) - np.sum(
			ldn_dm*lm)
		slope_estimate /= -np.sum(lm*lm) + 1/n_dm*np.sum(lm)**2

		# The MLE estimate on the norm
		norm_estimate = np.exp(1/n_dm*np.sum(ldn_dm-slope_estimate*lm))

		return slope_estimate, norm_estimate

	def two_halo_boost(self,z,z_lens,dz,lens_m200,r_max,r_min,n_quads=100):
		"""Calculates the boost from the two halo term of the host halo at
		the given redshift.

		Args:
			z (float): The redshift to calculate the boost at.
			z_len (float): The redshift of the main deflector
			dz (float): The thickness of the redshift slice to consider
			lens_m200 (float): The mass of the host lens in units of
				M_sun with mass definition 200c.
			r_max (float): The maximum radius to calculate the correlation
				function out to.
			r_min (float): The minimum radius to consider the correlation
				funciton to (to avoid overlap with substructure draws).
			n_quads (int): The number of points to use in averaging the
				two halo term of the redshift slice.

		Returns
			(float): The boost at the given redshift.
		"""
		# Get a range of z values we will average over.
		z_range = np.linspace(z,z+dz,n_quads)
		# Only consider the two point function in the regime where it is
		# large and we are outside the virial radius of the host halo.
		z_min = z_range[np.argmin(np.abs(z_range-z_lens))]
		r_cmv_min = np.abs(self.cosmo.comovingDistance(z_min,z_lens))
		if r_cmv_min >= r_max*self.cosmo.h:
			return 1

		r_cmv = np.abs(self.cosmo.comovingDistance(z_range,z_lens))
		# Don't try to calculate inside the virial radius.
		r_cmv = r_cmv[r_cmv>r_min*self.cosmo.h]

		# Get the two halo term in the slice
		xi_halo = self.cosmo.correlationFunction(r_cmv,z_lens)
		xi_halo *= bias.haloBias(lens_m200*self.cosmo.h,z_lens,mdef='200c',
			model='tinker10')
		return 1+np.mean(xi_halo)

	@functools.lru_cache(maxsize=1024)
	def cone_angle_to_radius(self,z,z_lens,z_source,cone_angle,
		angle_buffer=0.8):
		"""Returns the radius in kpc at the given redshift for the given cone
		angle.

		Args:
			z (float): The redshift at which to calculate the radius
			z_lens (float): The redshift of the main deflector
			z_source (float): The redshift of the source
			dz (float): The thickness of the redshift slice to consider
			cone_angle (float): The opening angle of the cone for the los
				realization in units of arcseconds.
			angle_buffer (float): A buffer for how small the second cone
				gets as it approaches the source. Should be between 0 and
				1 with 1 meaning that the second cone converges to a point
				at the source.

		Retruns:
			(float): The radius in units of physical kpc.
		"""
		# Get the conversion between the angular opening and kpc
		kpc_per_arcsecond = cosmology_utils.kpc_per_arcsecond(z,self.cosmo)
		r_los = kpc_per_arcsecond*cone_angle*0.5
		# If we're past the lens, shrink the light cone proportional to the
		# distance to the source.
		if z > z_lens:
			# Formula picked to match double cone in DG19
			scal_factor = angle_buffer
			scal_factor *= self.cosmo.comovingDistance(z_source)
			scal_factor /= self.cosmo.comovingDistance(z_lens,z_source)
			scal_factor *= self.cosmo.comovingDistance(z_lens,z)
			scal_factor /= self.cosmo.comovingDistance(z)
			r_los *= 1 - scal_factor
		return r_los

	@functools.lru_cache(maxsize=1024)
	def volume_element(self,z,z_lens,z_source,dz,cone_angle,angle_buffer=0.8):
		"""Returns the physical volume element at the given redshift

		Args:
			z (float): The redshift at which to calculate the volume element
			z_lens (float): The redshift of the main deflector
			z_source (float): The redshift of the source
			dz (float): The thickness of the redshift slice to consider
			cone_angle (float): The opening angle of the cone for the los
				realization in units of arcseconds.

		Retruns:
			(float): The physical volume element in kpc**3

		Notes:
			This depends on parameters like the cone angle, the redshift of
			the source, and the redshift of the lens.
		"""
		r_los = self.cone_angle_to_radius(z+dz/2,z_lens,z_source,cone_angle,
			angle_buffer=angle_buffer)

		# Get the thickness of our cone slice in physical units of kpc
		dz_in_kpc = self.cosmo.comovingDistance(z,z+dz)/self.cosmo.h
		dz_in_kpc /= (1+z)
		# Mpc to kpc
		dz_in_kpc *= 1000

		return dz_in_kpc * np.pi * r_los**2

	def draw_nfw_masses(self,z):
		"""Draws from the Sheth Tormen mass function with an additional
		correction for two point correlation with main lens.

		Args:
			z (float): The redshift at which to draw the masses

		Returns:
			(np.array): An array with the drawn masses in units of M_sun.

		"""
		# Pull the parameters we need from the input dictionaries
		# Units of M_sun
		lens_m200 = self.main_deflector_parameters['M200']
		z_lens = self.main_deflector_parameters['z_lens']
		z_source = self.source_parameters['z_source']
		dz = self.los_parameters['dz']
		# Units of arcsecond
		cone_angle = self.los_parameters['cone_angle']
		# Units of Mpc
		r_max = self.los_parameters['r_max']
		# Units of Mpc
		r_min = self.los_parameters['r_min']
		# Units of M_sun
		m_min = self.los_parameters['m_min']
		# Units of M_sun
		m_max = self.los_parameters['m_max']
		delta_los = max(0, self.los_parameters['delta_los'])
		# Get the parameters of the power law fit to the Sheth Tormen mass
		# function
		pl_slope, pl_norm = self.power_law_dn_dm(z+dz/2,m_min,m_max)

		# Scale the norm by the total volume and the two point correlation.
		dV = self.volume_element(z,z_lens,z_source,dz,cone_angle)
		halo_boost = self.two_halo_boost(z,z_lens,dz,lens_m200,r_max,r_min)
		pl_norm *= dV * halo_boost * delta_los

		# Draw from our power law and return the masses.
		masses = power_law.power_law_draw(m_min,m_max,pl_slope,pl_norm)
		return masses

	def sample_los_pos(self,z,n_los):
		"""Draws the positions for the line of sight substructure at the given
		redshift.

		Args:
			z (float): The redshift to place the los halos at.
			n_los (int): The number of los halos to draw position for

		Returns:
			(np.array): A n_los x 2 array giving the x,y position of the
			line of sight structure in units of kpc.
		"""
		# Pull the parameters we need from the input dictionaries
		# Units of M_sun
		z_lens = self.main_deflector_parameters['z_lens']
		z_source = self.source_parameters['z_source']
		dz = self.los_parameters['dz']
		# Units of arcsecond
		cone_angle = self.los_parameters['cone_angle']

		r_los = self.cone_angle_to_radius(z+dz/2,z_lens,z_source,cone_angle)
		# Draw the radii of the los halos.
		r_draws = r_los*np.sqrt(np.random.rand(n_los))
		theta = 2*np.pi*np.random.rand(n_los)

		# Create an array for the coordinates
		cart_pos = np.zeros((n_los,2))

		cart_pos[:,0] = r_draws * np.cos(theta)
		cart_pos[:,1] = r_draws * np.sin(theta)

		return cart_pos

	def mass_concentration(self,z,m_200,scatter_mult=1.0):
		"""Returns the concentration of halos at a certain mass given the
		parameterization of DG_19.

		Args:
			z (float): The redshift of the nfw halos
			m_200 (np.array): array of M_200 of the nfw halo units of M_sun
			scatter_mult (float): an additional scaling to the scatter. Likely
				only useful for los rendering to force scatter to 0.

		Returns:
			(np.array): The concentration for each halo.
		"""
		# Get the concentration parameters
		c_0 = self.los_parameters['c_0']
		zeta = self.los_parameters['conc_zeta']
		beta = self.los_parameters['conc_beta']
		m_ref = self.los_parameters['conc_m_ref']
		dex_scatter = self.los_parameters['dex_scatter']*scatter_mult

		# The peak calculation is done by colossus. The cosmology must have
		# already been set. Note these functions expect M_sun/h units (which
		# you get by multiplying by h
		# https://www.astro.ljmu.ac.uk/~ikb/research/h-units.html)
		h = self.cosmo.h
		peak_heights = peaks.peakHeight(m_200*h,z)
		peak_height_ref = peaks.peakHeight(m_ref*h,0)

		# Now get the concentrations and add scatter
		concentrations = c_0*(1+z)**(zeta)*(peak_heights/peak_height_ref)**(
			-beta)
		if isinstance(concentrations,np.ndarray):
			conc_scatter = np.random.randn(len(concentrations))*dex_scatter
		elif isinstance(concentrations,float):
			conc_scatter = np.random.randn()*dex_scatter
		concentrations = 10**(np.log10(concentrations)+conc_scatter)

		return concentrations

	def convert_to_lenstronomy(self,z,z_masses,z_cart_pos):
		"""Converts the subhalo masses and position to truncated NFW profiles
		for lenstronomy

		Args:
			z (float): The redshift for each of the halos
			z_masses (np.array): The masses of each of the halos that
				were drawn
			z_cart_pos (np.array): A n_los x 2D array of the position of the
				halos that were drawn
		Returns:
			([string,...],[dict,...]): A tuple containing the list of models
			and the list of kwargs for the truncated NFWs.
		"""
		z_source = self.source_parameters['z_source']
		# First, draw a concentration for all our LOS structure from our mass
		# concentration relation
		concentration = self.mass_concentration(z,z_masses)

		# Now convert our mass and concentration into the lenstronomy
		# parameters
		z_r_200 = nfw_functions.r_200_from_m(z_masses,z,self.cosmo)
		z_r_scale = z_r_200/concentration
		z_rho_nfw = nfw_functions.rho_nfw_from_m_c(z_masses,concentration,
			self.cosmo,r_scale=z_r_scale)

		# Convert to lenstronomy units
		z_r_scale_ang, alpha_Rs = nfw_functions.convert_to_lenstronomy_NFW(
			z_r_scale,z,z_rho_nfw,z_source,self.cosmo)
		kpc_per_arcsecond = cosmology_utils.kpc_per_arcsecond(z,self.cosmo)
		cart_pos_ang = z_cart_pos / np.expand_dims(kpc_per_arcsecond,-1)

		# Populate the parameters for each lens
		model_list = []
		kwargs_list = []

		for i in range(len(z_masses)):
			model_list.append('NFW')
			kwargs_list.append({'alpha_Rs':alpha_Rs[i], 'Rs':z_r_scale_ang[i],
				'center_x':cart_pos_ang[i,0],'center_y':cart_pos_ang[i,1]})

		return (model_list,kwargs_list)

	def draw_los(self):
		"""Draws masses, concentrations,and positions for the los substructure
		of a main lens halo.

		Returns:
			(tuple): A tuple of three lists: the first is the profile type for
			each los halo returned, the second is the lenstronomy kwargs
			for that halo, and the third is a list of redshift values for
			each profile.
		Notes:
			The returned lens model list includes terms to correct for
			the average deflection angle introduced from the los halos.
		"""
		# Distribute line of sight substructure according to
		# https://arxiv.org/pdf/1909.02573.pdf. This also includes a
		# correction for the average deflection angle introduced by
		# the addition of the substructure.
		los_model_list = []
		los_kwargs_list = []
		los_z_list = []

		# Pull the paramters we need
		z_min = self.los_parameters['z_min']
		z_source = self.source_parameters['z_source']
		dz = self.los_parameters['dz']

		# Add halos from the starting reshift to the source redshift.
		# Note most of the calculations are done at z + dz/2, so you
		# want to stop at z_source-dz.
		z_range = np.arange(z_min,z_source-dz,dz)
		# Round the z_range to improve caching hits.
		z_range = list(np.round(z_range,2))

		# Iterate through each z and add the halos.
		for z in z_range:
			# Draw the masses and positions at this redshift from our
			# model
			z_masses = self.draw_nfw_masses(z)
			# Don't add anything to the model if no masses were drawn
			if z_masses.size == 0:
				continue
			z_cart_pos = self.sample_los_pos(z,len(z_masses))
			# Convert the mass and positions to lenstronomy models
			# and kwargs and append to our lists.
			model_list, kwargs_list = self.convert_to_lenstronomy(
				z,z_masses,z_cart_pos)
			los_model_list += model_list
			los_kwargs_list += kwargs_list

			# TODO - correction for average line of sight

			los_z_list += [z+dz/2]*len(model_list)

		return (los_model_list, los_kwargs_list, los_z_list)

	def calculate_average_alpha(self,num_pix):
		""" Calculates the average deflection maps from the los at each
		redshift specified by the los parameters and returns corresponding
		lenstronomy objects.

		Args:
			num_pix (int): The number of pixels to sample for our
				interpolation maps.

		Returns:
			(tuple): A tuple of two lists: the first is the interpolation
			profile type for each redshift slice and the second is the
			lenstronomy kwargs for that profile.

		Notes:
			The average los deflection angles of the lenstronomy objects will
			be the negative of the average (since we want to subtract the
			average effect not add it). Pixel scale will be set such that
			at each redshift a box of 5*r_los is captured.
		"""
		# Pull the parameters we need.
		z_min = self.los_parameters['z_min']
		z_source = self.source_parameters['z_source']
		z_lens = self.main_deflector_parameters['z_lens']
		dz = self.los_parameters['dz']
		dz *= self.los_parameters['alpha_dz_factor']
		delta_los = max(0, self.los_parameters['delta_los'])
		cone_angle = self.los_parameters['cone_angle']
		m_min = self.los_parameters['m_min']
		# Units of M_sun
		m_max = self.los_parameters['m_max']
		z_range = np.arange(z_min,z_source-dz,dz)
		# Round the z_range to improve caching hits. Add the dz/2 shift that
		# gets output by draw_los.
		z_range = list(np.round(z_range,2)+dz/2)

		# The lists where we'll store the lenstornomy variables
		interp_model_list = []
		interp_kwargs_list = []
		interp_z_list = []

		for zi, z in enumerate(z_range):
			# First we make the grid on which we'll conduct the interpolation.
			r_los = self.cone_angle_to_radius(z+dz/2,z_lens,z_source,
				cone_angle)
			kpc_per_arcsecond = cosmology_utils.kpc_per_arcsecond(z,self.cosmo)
			r_los = r_los/kpc_per_arcsecond

			# Set the pixel scale to capture a box of dimension 5*r_los.
			pixel_scale = 5*r_los/num_pix
			x_grid, y_grid = util.make_grid(numPix=num_pix,
				deltapix=pixel_scale)
			x_axes, y_axes = util.get_axes(x_grid, y_grid)

			# Then we create the 2d rendering disk we want to convolve our
			# NFW with
			disk_bool = np.zeros(x_grid.shape)
			disk_bool[np.sqrt(x_grid**2+y_grid**2)<r_los] = 1
			disk_bool = util.array2image(disk_bool)

			# Next we calculate the NFW parameters for our NFW of average
			# mass.
			pl_slope, pl_norm = self.power_law_dn_dm(z+dz/2,m_min,m_max)
			m_average = (power_law.power_law_integrate(m_min,m_max,pl_slope+1)/
				power_law.power_law_integrate(m_min,m_max,pl_slope))
			c_average = self.mass_concentration(z,m_average,scatter_mult=0.0)

			# Convert our parameters to the lenstronomy definition
			r_200_avg = nfw_functions.r_200_from_m(m_average,z,self.cosmo)
			r_scale_avg = r_200_avg/c_average
			rho_nfw_avg = nfw_functions.rho_nfw_from_m_c(m_average,c_average,
				self.cosmo,r_scale=r_scale_avg)
			r_scale_ang_avg, alpha_Rs_avg = (
				nfw_functions.convert_to_lenstronomy_NFW(r_scale_avg,z,
				rho_nfw_avg,z_source,self.cosmo))

			# Get our deflection angle and potential for this NFW profile
			ax,ay = NFW().derivatives(x_grid,y_grid,r_scale_ang_avg,
				alpha_Rs_avg)
			a = NFW().function(x_grid,y_grid,r_scale_ang_avg,alpha_Rs_avg)
			ax = util.array2image(ax)
			ay = util.array2image(ay)
			a = util.array2image(a)

			# Convolve our deflection angles and potential with the disk
			ax_conv = fftconvolve(ax,disk_bool,mode='same')
			ay_conv = fftconvolve(ay,disk_bool,mode='same')
			a_conv = fftconvolve(a,disk_bool,mode='same')

			# Finally, we just need to calculate how many of our
			# m_average NFW we expect per pixel. So far our convolution
			# is the deflection angle assuming 1 NFW / pixel.
			# First we calculate the number of m_avg nfw per kpc^3
			m_total = power_law.power_law_integrate(m_min,m_max,pl_slope+1)
			m_total *= pl_norm * delta_los
			n_total = m_total/m_average

			# Now we multiply by the length of our redshift slice to get per
			# kpc^2 and then multiply by kpc_per_arcsecond to get per
			# arcsecond^2
			dz_in_kpc = self.cosmo.comovingDistance(z-dz/2,z+dz/2)/self.cosmo.h
			dz_in_kpc /= (1+z-dz/2)
			# Mpc to kpc
			dz_in_kpc *= 1000
			n_total *= dz_in_kpc
			n_total *= kpc_per_arcsecond**2

			# Final convert from per arcsecond^2 to per pixel
			n_total *= pixel_scale**2

			ax_conv *= n_total
			ay_conv *= n_total
			a_conv *= n_total

			# And now we can return the parameters of our interpol
			# profile
			interp_model_list.append('INTERPOL')
			interp_kwargs_list.append({'grid_interp_x':x_axes,
				'grid_interp_y':y_axes, 'f_x':-ax_conv,
				'f_y':-ay_conv,'f_':-a_conv})
			interp_z_list.append(z)

		return (interp_model_list,interp_kwargs_list,interp_z_list)



from paltas.Configs.xxxx_yyyy.config_val import *

config_dict['subhalo']['parameters']['sigma_sub'] = norm(loc=3.8e-3,
	scale=1.5e-4).rvs
config_dict['main_deflector']['parameters']['gamma'] = truncnorm(-195.8,
	np.inf,loc=1.958,scale=0.01).rvs
config_dict['main_deflector']['parameters']['theta_E'] = truncnorm(-73.8,
	np.inf,loc=1.108,scale=0.015).rvs
config_dict['main_deflector']['parameters']['e1'] = norm(loc=-0.034,
	scale=0.01).rvs
config_dict['main_deflector']['parameters']['e2'] = norm(loc=-0.032,
	scale=0.01).rvs
config_dict['main_deflector']['parameters']['center_x'] = norm(loc=0.027,
	scale=0.016).rvs
config_dict['main_deflector']['parameters']['center_y'] = norm(loc=-0.070,
	scale=0.016).rvs
config_dict['main_deflector']['parameters']['gamma1'] = norm(loc=-0.009,
	scale=0.005).rvs
config_dict['main_deflector']['parameters']['gamma2'] = norm(loc=0.003,
	scale=0.005).rvs



# -*- coding: utf-8 -*-
"""
Initialize and train a neural posterior estimator on a strong lensing image
dataset generated by paltas.
"""
import argparse, os, sys, glob, math
from importlib import import_module
import tensorflow as tf
from . import dataset_generation, loss_functions, conv_models
from tensorflow.keras.callbacks import TensorBoard, ModelCheckpoint
from tensorflow.keras import optimizers
import pandas as pd


def parse_args():
	"""Parse the input arguments by the user

	Returns:
		(argparse.Namespace): An instance of the Namespace object with the
		users provided values.

	"""
	# Initialize the parser and the possible inputs
	parser = argparse.ArgumentParser()
	parser.add_argument('training_config', help='Path to configuration for '
		+'training the model.')
	parser.add_argument('--tensorboard_dir', default=None, type=str,
		dest='tensorboard_dir', help='Optional path to save a tensorboard ' +
		'output to')
	args = parser.parse_args()
	return args


def main():
	"""The main script to run from the command line.
	"""
	# Get the user provided arguments
	args = parse_args()

	# Get the training parameters from the provided .py file
	config_dir, config_file = os.path.split(os.path.abspath(
		args.training_config))
	sys.path.insert(0, config_dir)
	config_name, _ = os.path.splitext(config_file)
	config_module = import_module(config_name)

	# Grab the arguments passed in by the user
	# The size of each training batch
	batch_size = config_module.batch_size
	# The number of epochs to train for
	n_epochs = config_module.n_epochs
	# The size of the images in the training set
	img_size = config_module.img_size
	# A list of the paths to the TFRecords for the training images
	tfr_train_paths = config_module.tfr_train_paths
	# The path to the TFRecord for the validation images
	tfr_val_path = config_module.tfr_val_path
	# The list of learning parameters to use
	learning_params = config_module.learning_params
	# Which parameters to put in log space
	if hasattr(config_module,'log_learning_params'):
		log_learning_params = config_module.log_learning_params
	else:
		log_learning_params = []
	num_params = len(learning_params+log_learning_params)
	# Which parameters to consider flipping
	if hasattr(config_module,'flip_pairs'):
		flip_pairs = config_module.flip_pairs
	else:
		flip_pairs = None
	# Which parameters to weight
	if hasattr(config_module,'weight_terms'):
		weight_terms = config_module.weight_terms
	else:
		weight_terms = None
	# Whether to train the full model or only the head
	if hasattr(config_module,'train_only_head'):
		train_only_head = config_module.train_only_head
	else:
		train_only_head = False
	# A list to the paths of the fodlers containing the npy images
	# for training
	npy_folders_train = config_module.npy_folders_train
	# Number of steps per epoch is number of examples over the batch size
	n_npy_files = 0
	for npy_folder in npy_folders_train:
		n_npy_files += len(glob.glob(os.path.join(npy_folder,'image_*.npy')))
	steps_per_epoch = n_npy_files//batch_size
	# The path to the fodler containing the npy images
	# for validation
	npy_folder_val = config_module.npy_folder_val
	# Get the number of validation files as well
	n_val_npy = len(glob.glob(os.path.join(npy_folder_val,'image_*.npy')))
	# A list of the paths to the training metadata
	metadata_paths_train = config_module.metadata_paths_train
	# The path to the validation metadata
	metadata_path_val = config_module.metadata_path_val
	# The path to the csv file to read from / write to for normalization
	# of learning parameters.
	input_norm_path = config_module.input_norm_path
	# The detector kwargs to use for on-the-fly noise generation
	kwargs_detector = config_module.kwargs_detector
	# Whether or not to normalize the images by the standard deviation
	norm_images = config_module.norm_images
	# A string with which loss function to use.
	loss_function = config_module.loss_function
	# A string specifying which model to use
	model_type = config_module.model_type
	# A string specifying which optimizer to use
	optimizer_string = config_module.optimizer
	# Where to save the model weights and where to load the initial weights
	model_weights_init = config_module.model_weights_init
	model_weights = config_module.model_weights
	# The learning rate for the model
	learning_rate = config_module.learning_rate
	# Whether or not to apply a random rotation to the input image
	random_rotation = config_module.random_rotation

	# Set the random seed for our network
	# tf.random.set_seed(random_seed)

	# Check for tf records for train and validation and prepare them
	# if needed
	print('Checking for training data.')
	for i, tf_path in enumerate(tfr_train_paths):
		if not os.path.exists(tf_path):
			print('Generating new TFRecord at %s'%(tf_path))
			dataset_generation.generate_tf_record(npy_folders_train[i],
				learning_params+log_learning_params,metadata_paths_train[i],
				tf_path)
		else:
			print('TFRecord found at %s'%(tf_path))

	print('Checking for validation data.')
	if not os.path.exists(tfr_val_path):
		print('Generating new TFRecord at %s'%(tfr_val_path))
		dataset_generation.generate_tf_record(npy_folder_val,
			learning_params+log_learning_params,metadata_path_val,tfr_val_path)
	else:
		print('TFRecord found at %s'%(tfr_val_path))

	# Normalize outputs if requested
	if input_norm_path is not None:
		print('Checking for normalization csv')
		metadata = pd.read_csv(metadata_paths_train[0])
		dataset_generation.normalize_outputs(metadata,learning_params,
			input_norm_path,log_learning_params=log_learning_params)

	# If no random rotations are required, the best tool is a tf dataset.
	# Otherwise, it's better to take the output of the tf dataset and then
	# run it through a generator that will do the rotations on the batches
	# for us.
	if random_rotation:
		# Get a generator object that returns rotated images and parameters.
		tf_dataset_t = dataset_generation.generate_rotations_dataset(
			tfr_train_paths,learning_params,batch_size,n_epochs,
			norm_images=norm_images,input_norm_path=input_norm_path,
			kwargs_detector=kwargs_detector,
			log_learning_params=log_learning_params)
	else:
		# Turn our tf records into tf datasets for training and validation
		tf_dataset_t = dataset_generation.generate_tf_dataset(tfr_train_paths,
			learning_params,batch_size,n_epochs,norm_images=norm_images,
			input_norm_path=input_norm_path,kwargs_detector=kwargs_detector,
			log_learning_params=log_learning_params)
	# We shouldn't be adding random noise to validation images. They should
	# be generated with noise
	if kwargs_detector is not None:
		print('Make sure your validation images already have noise! Noise ' +
			'will not be added on the fly for validation.')
	tf_dataset_v = dataset_generation.generate_tf_dataset(tfr_val_path,
		learning_params,min(batch_size,n_val_npy),1,norm_images=norm_images,
		input_norm_path=input_norm_path,kwargs_detector=None,
		log_learning_params=log_learning_params)

	print('Initializing the model')

	# Load the loss function
	if loss_function == 'mse':
		num_outputs = num_params
		loss = loss_functions.MSELoss(num_params,flip_pairs,weight_terms).loss
	elif loss_function == 'diag':
		num_outputs = num_params*2
		loss = loss_functions.DiagonalCovarianceLoss(num_params,
			flip_pairs,weight_terms).loss
	elif loss_function == 'full':
		num_outputs = num_params + int(num_params*(num_params+1)/2)
		loss = loss_functions.FullCovarianceLoss(num_params,flip_pairs,
			weight_terms).loss
	else:
		raise ValueError('%s loss not in the list of supported losses'%(
			loss_function))

	# Load the model
	if model_type == 'xresnet34':
		model = conv_models.build_xresnet34(img_size,num_outputs,
			train_only_head=train_only_head)
	else:
		raise ValueError('%s model not in the list of supported models'%(
			model_type))

	# Use learning rate decay for optimal learning
	lr_schedule = tf.keras.optimizers.schedules.ExponentialDecay(
		learning_rate,decay_steps=steps_per_epoch,decay_rate=0.98,
		staircase=True)

	# Use the desired optimizer
	opt = getattr(optimizers,optimizer_string)(learning_rate=lr_schedule)

	# Compile our model
	model.compile(loss=loss,optimizer=opt,metrics=[loss])

	print('Is model built: ' + str(model.built))

	try:
		model.load_weights(model_weights_init)
		print('Loaded weights %s'%(model_weights_init))
	except:
		print('No weights found. Saving new weights to %s'%(model_weights))

	# Set up the callbacks for our training
	callbacks = []
	# If requested, save our model progress to a tensorboard directory
	if args.tensorboard_dir is not None:
		tensorboard = TensorBoard(log_dir=args.tensorboard_dir,
			update_freq='batch')
		callbacks.append(tensorboard)
	# Save the model weights as long as the validation loss is decreasing
	modelcheckpoint = ModelCheckpoint(model_weights,monitor='val_loss',
		save_best_only=False,save_freq='epoch')
	callbacks.append(modelcheckpoint)

	# TODO add validation data.
	model.fit(tf_dataset_t,callbacks=callbacks,epochs=n_epochs,
		steps_per_epoch=steps_per_epoch,validation_data=tf_dataset_v,
		validation_steps=int(math.ceil(n_val_npy/batch_size)))


if __name__ == '__main__':
	main()



# -*- coding: utf-8 -*-
"""
Turn real galaxies into Lenstronomy inputs.

This module contains the default class for transforming the objects of a
source catalog into sources to be passed to lenstronomy.
"""
import numpy as np
from .source_base import SourceBase


class GalaxyCatalog(SourceBase):
	"""Base class for turning real galaxy images into Lenstronomy inputs.

	Args:
		cosmology_parameters (str,dict, or colossus.cosmology.Cosmology):
			Either a name of colossus cosmology, a dict with 'cosmology name':
			name of colossus cosmology, an instance of colussus cosmology, or a
			dict with H0 and Om0 ( other parameters will be set to defaults).
		source_parameters (dict): A dictionary containing all the parameters
			needed to draw sources (in this case random_rotation).
	"""
	required_parameters = ('random_rotation','output_ab_zeropoint',
		'z_source','center_x','center_y')
	# This parameter must be set by class inheriting GalaxyCatalog
	ab_zeropoint = None

	def __len__(self):
		"""Returns the length of the catalog"""
		raise NotImplementedError

	def check_parameterization(self, required_params):
		""" Check that all the required parameters are present in the
		source_parameters. Also checks ab_zeropoint is set.

		Args:
			required_params ([str,...]): A list of strings containing the
				required parameters.
		"""
		# Run the base check
		super().check_parameterization(required_params)
		# Check ab magnitude zeropoint has been set
		if (self.__class__.__name__ != 'GalaxyCatalog' and
			self.__class__.ab_zeropoint is None):
			raise ValueError('ab_zeropoint must be set by class inheriting '+
				'GalaxyCatalog.')

	def image_and_metadata(self, catalog_i):
		"""Returns the image array and metadata for one galaxy

		Args:
			catalog_i (int): The catalog index

		Returns
			([np.array, np.void]) A numpy array containing the image
			metadata and a numpy void type that acts as a dictionary with
			the metadata.
		"""
		raise NotImplementedError

	def iter_lightmodel_kwargs_samples(self,n_galaxies,z_new,
		**selection_kwargs):
		"""Yields dicts of lenstronomy LightModel kwargs for n_galaxies,
		placed at redshift z_new

		Args:
			n_galaxies (int): Number of galaxies to draw
			z_new (float): Redshift to place galaxies at
			**selection_kwargs: Kwargs that can be passed to the
				sample_indices method.

		Returns:
			(generator): A generator that can be iterated over to give
			lenstronomy model lists and kwargs.
		"""
		for catalog_i in self.sample_indices(n_galaxies,**selection_kwargs):
			yield self.draw_source(catalog_i)[:2]

	def iter_image_and_metadata(self, message=''):
		"""Yields the image array and metadata for all of the images
		in the catalog.

		Args:
			message (str): If the iterator uses tqdm, this message
				will be displayed.

		Returns:
			(generator): A generator that can be iterated over to give
			lenstronomy kwargs.
		"""
		for catalog_i in range(len(self)):
			yield self.image_and_metadata(catalog_i)

	def sample_indices(self, n_galaxies):
		"""Returns n_galaxies array of ints, catalog indices to sample

		Args:
			n_galaxies (int): Number of indices to return

		Returns:
			(np.array): Array of ints to sample
		"""
		return np.random.randint(0, len(self), size=n_galaxies)

	def fill_catalog_i_phi_defaults(self, catalog_i=None, phi=None):
		"""Return catalog index and source rotation angle.

		Args:
			catalog_i (int): Index of image in catalog
				If not provided or None, will be sampled from catalog.
			phi (float): Rotation to apply to the image.
				If not provided or None, will either randomize or use 0,
				depending on source_parameters['random_rotation'].
		"""
		# If no index is provided pick one at random
		if catalog_i is None:
			catalog_i = self.sample_indices(1).item()
		# If no rotation is provided, pick one at random or use original
		# orientation
		if phi is None:
			if self.source_parameters['random_rotation']:
				phi = self.draw_phi()
			else:
				phi = 0
		return catalog_i, phi

	def draw_source(self, catalog_i=None, phi=None):
		"""Creates lenstronomy lightmodel kwargs from a catalog image.

		Args:
			catalog_i (int): Index of image in catalog
			z_new (float): Redshift to place image at
			phi (float): Rotation to apply to the image.
				If not provided, randomize or use 0, depending on
				source_parameters['random_rotation']

		Returns:
			(list,list,list): A list containing the model ['INTERPOL'],
			the kwargs for an instance of the class
			lenstronomy.LightModel.Profiles.interpolation.Interpol,
			and the redshift of the model.

		Notes:
			If not catalog_i is provided, one that meets the cuts will be
			selected at random.
		"""
		catalog_i, phi = self.fill_catalog_i_phi_defaults(catalog_i, phi)
		img, metadata = self.image_and_metadata(catalog_i)
		pixel_width = metadata['pixel_width']
		z_new = self.source_parameters['z_source']

		# With this, lenstronomy will preserve the scale/units of
		# the input image (in a configuration without lensing,
		# same pixel widths, etc.)
		img = img / pixel_width**2

		# Take into account the difference in the magnitude zeropoints
		# of the input survey and the output survey. Note this doesn't
		# take into account the color of the object!
		img *= 10**((self.source_parameters['output_ab_zeropoint']-
			self.__class__.ab_zeropoint)/2.5)

		pixel_width *= self.z_scale_factor(metadata['z'], z_new)

		# Convert to kwargs for lenstronomy
		return (
			['INTERPOL'],
			[dict(
				image=img,
				center_x=self.source_parameters['center_x'],
				center_y=self.source_parameters['center_y'],
				phi_G=phi,
				scale=pixel_width)],
			[z_new])

	@staticmethod
	def draw_phi():
		"""Draws a random rotation angle for the interpolation of the source.

		Returns:
			(float): The new angle to use in the interpolation class.
		"""
		return np.random.rand() * 2 * np.pi

	def z_scale_factor(self, z_old, z_new):
		"""Return multiplication factor for object/pixel size for moving its
		redshift from z_old to z_new.

		Args:
			z_old (float): The original redshift of the object.
			z_new (float): The redshift the object will be placed at.

		Returns:
			(float): The multiplicative pixel size.
		"""
		# Pixel length ~ angular diameter distance
		# (colossus uses funny /h units, but for ratios it
		#  fortunately doesn't matter)
		return (
			self.cosmo.angularDiameterDistance(z_old)
			/ self.cosmo.angularDiameterDistance(z_new))



# -*- coding: utf-8 -*-
"""
Construct the loss functions for use in NN training.

This module contains classes that can be used to initialize tensorflow loss
functions for NN training on the strong lensing problem.
"""

import tensorflow as tf
import numpy as np
import itertools


class BaseLoss():
	"""	A base class for the loss functions.

	Args:
		num_params (int): The number of parameters to predict.
		flip_pairs ([[int,...],...]): A list of lists. Each list contains
			the index of parameters that when flipped together return an
			equivalent lens model.
		weight_terms ([[int,...],...]): A list of lists. Each list contains
			the index of a parameter and the value by which you would like
			that parameters loss to be weighted.

	Notes:
		If multiple lists are provided, all possible combinations of
		flips will be considered. For example, if flip_pairs is [[0,1],[2,3]]
		then flipping 0,1,2,3 all at the same time will also be considered.
	"""

	def __init__(self, num_params, flip_pairs=None, weight_terms=None):
		# Save the flip pairs for later use
		self.flip_pairs = flip_pairs
		self.num_params = num_params
		self.weight_terms = weight_terms

		# Create a vector that will be used to rescale the loss
		self.weight_vector = np.ones(num_params)
		if weight_terms is not None:
			for wt in self.weight_terms:
				self.weight_vector[wt[0]] *= wt[1]
		self.weight_vector = tf.constant(self.weight_vector,dtype=tf.float32)

		# Now for each flip pair (including no flip) we will add a flip
		# matrix to our list.
		self.flip_mat_list = [tf.linalg.diag(tf.ones((self.num_params),
			dtype=tf.float32))]
		if self.flip_pairs is not None:
			# Create a flip matrix for every possible unique combination
			for flip_i in range(1,len(self.flip_pairs)+1):
				for comb in itertools.combinations(self.flip_pairs,flip_i):
					flip_list = list(itertools.chain.from_iterable(comb))
					const_initializer = np.ones(self.num_params)
					const_initializer[flip_list] = -1
					self.flip_mat_list.append(tf.linalg.diag(tf.constant(
						const_initializer,dtype=tf.float32)))

	def convert_output(self,output):
		"""	Convert the model outputs into the different components of the
		loss function

		Args:
			output (tf.Tensor): The predicted values of the lensing parameters

		Returns:
			([tf.Tensor,...]): The list of tf.Tensor objects that correspond to
			the components of the loss function.
		"""

		raise NotImplementedError('convert_output has not been defined.')

	def draw_samples(self,output,n_samps):
		"""	Draw samples from the predicted posterior defined by the networks
		outputs.

		Args:
			output (tf.Tensor): The predicted values of the lensing parameters
			n_samps (int): The number of samples to draw from each posterior.

		Returns:
			(np.array): An array with shape (n_samps,batch_size,n_params)
			containing draws from the posterior of each lens defined in
			the output.
		"""

		raise NotImplementedError('convert_output has not been defined.')


class MSELoss(BaseLoss):
	""" MSE loss that includes parameter flipping

	Args:
		num_params (int): The number of parameters to predict.
		flip_pairs ([[int,...],...]): A list of lists. Each list contains
			the index of parameters that when flipped together return an
			equivalent lens model.
		weight_terms ([[int,...],...]): A list of lists. Each list contains
			the index of a parameter and the value by which you would like
			that parameters loss to be weighted.

	Notes:
		If multiple lists are provided, all possible combinations of
		flips will be considered. For example, if flip_pairs is [[0,1],[2,3]]
		then flipping 0,1,2,3 all at the same time will also be considered.
	"""

	def convert_output(self,output):
		"""	Convert the model outputs into the different components of the
		loss function

		Args:
			output (tf.Tensor): The predicted values of the lensing parameters

		Returns:
			([tf.Tensor,...]): The list of tf.Tensor objects that correspond to
			the components of the loss function.
		"""
		# Get the mle value from the output.
		y_pred, _ = tf.split(output,num_or_size_splits=(self.num_params,-1),
			axis=-1)
		return y_pred

	def draw_samples(self,output,n_samps):
		"""	Draw samples from the predicted posterior defined by the networks
		outputs.

		Args:
			output (tf.Tensor): The predicted values of the lensing parameters
			n_samps (int): The number of samples to draw from each posterior.

		Returns:
			(np.array): An array with shape (n_samps,batch_size,n_params)
			containing draws from the posterior of each lens defined in
			the output.
		"""
		# The posterior is just a delta function.
		y_pred = self.convert_output(output).numpy()
		return np.repeat(np.expand_dims(y_pred,axis=0),n_samps,axis=0)

	def loss(self,y_true,output):
		""" Returns the MSE loss of the predicted parameters.

		Args:
			y_true (tf.Tensor): The true values of the parameters
			output (tf.Tensor): The predicted values of the lensing parameters

		Returns:
			(tf.Tensor): The loss function as a tf.Tensor.

		Notes:
			Can be used on networks that output more than just the maximum
			likelihood estimate, so long as the first num_params values
			outputted are the MLE.
		"""
		y_pred = self.convert_output(output)

		# Build a loss list and take the minimum value according to the
		# flip pairs.
		loss_list = []
		for flip_mat in self.flip_mat_list:
			loss_list.append(tf.reduce_mean(tf.square(
				tf.matmul(y_pred,flip_mat)-y_true)*self.weight_vector,axis=-1))
		loss_stack = tf.stack(loss_list,axis=-1)
		return tf.reduce_min(loss_stack,axis=-1)


class DiagonalCovarianceLoss(BaseLoss):
	""" Diagonal covariance loss that includes parameter flipping

	Args:
		num_params (int): The number of parameters to predict.
		flip_pairs ([[int,...],...]): A list of lists. Each list contains
			the index of parameters that when flipped together return an
			equivalent lens model.

		Notes:
			If multiple lists are provided, all possible combinations of
			flips will be considered. For example, if flip_pairs is
			[[0,1],[2,3]] then flipping 0,1,2,3 all at the same time will
			also be considered.
	"""

	def convert_output(self,output):
		"""	Convert the model outputs into the different components of the
		loss function

		Args:
			output (tf.Tensor): The predicted values of the lensing parameters

		Returns:
			([tf.Tensor,...]): The list of tf.Tensor objects that correspond to
			the components of the loss function.
		"""
		# Get both the mle value and the standard deviation.
		y_pred, log_var_pred = tf.split(output,num_or_size_splits=2,axis=-1)
		return y_pred, log_var_pred

	def draw_samples(self,output,n_samps):
		"""	Draw samples from the predicted posterior defined by the networks
		outputs.

		Args:
			output (tf.Tensor): The predicted values of the lensing parameters
			n_samps (int): The number of samples to draw from each posterior.

		Returns:
			(np.array): An array with shape (n_samps,batch_size,n_params)
			containing draws from the posterior of each lens defined in
			the output.
		"""
		# The posterior is the mean with random noise.
		y_pred, log_var_pred = self.convert_output(output)
		y_pred = y_pred.numpy()
		log_var_pred = log_var_pred.numpy()

		predict_samps = y_pred + np.random.randn(n_samps,y_pred.shape[0],
			y_pred.shape[1])*np.exp(log_var_pred/2)
		return predict_samps

	@staticmethod
	def log_gauss_diag(y_true,y_pred,log_var_pred):
		""" Return the negative log posterior of a Gaussian with diagonal
		covariance matrix

		Args:
			y_true (tf.Tensor): The true values of the parameters
			y_pred (tf.Tensor): The predicted value of the parameters
			log_var_pred (tf.Tensor): The predicted diagonal entries of the
				covariance. Note that log_var_pred is assumed to be the log of
				the covariance matrix values.

		Returns:
			(tf.Tensor): The TF graph for calculating the nlp

		Notes:
			This loss does not include the constant factor of 1/(2*pi)^(d/2).
		"""
		return 0.5*tf.reduce_sum(tf.multiply(tf.square(y_pred-y_true),
			tf.exp(-log_var_pred)),axis=-1) + 0.5*tf.reduce_sum(
			log_var_pred,axis=-1)

	def loss(self,y_true,output):
		""" Returns the loss of the predicted parameters.

		Args:
			y_true (tf.Tensor): The true values of the parameters
			output (tf.Tensor): The predicted values of the lensing
				parameters. This should include 2*self.num_params parameters
				to account for the diagonal entries of our covariance matrix.
				Covariance matrix values are assumed to be in log space.

		Returns:
			(tf.Tensor): The loss function as a tf.Tensor.
		"""
		y_pred, log_var_pred = self.convert_output(output)

		# Add each possible flip to the loss list. We will then take the
		# minimum.
		loss_list = []
		for flip_mat in self.flip_mat_list:
			loss_list.append(self.log_gauss_diag(y_true,
				tf.matmul(y_pred,flip_mat),log_var_pred))
		loss_stack = tf.stack(loss_list,axis=-1)
		return tf.reduce_min(loss_stack,axis=-1)


class FullCovarianceLoss(BaseLoss):
	""" Full covariance loss that includes parameter flipping

	Args:
		num_params (int): The number of parameters to predict.
		flip_pairs ([[int,...],...]): A list of lists. Each list contains
			the index of parameters that when flipped together return an
			equivalent lens model.

		Notes:
			If multiple lists are provided, all possible combinations of
			flips will be considered. For example, if flip_pairs is
			[[0,1],[2,3]] then flipping 0,1,2,3 all at the same time will
			also be considered.
	"""
	def __init__(self, num_params, flip_pairs=None, weight_terms=None):
		super().__init__(num_params,flip_pairs=flip_pairs,
			weight_terms=weight_terms)

		# Calculate the split list for lower traingular matrix
		self.split_list = []
		for i in range(1,num_params+1):
			self.split_list += [i]

	def convert_output(self,output):
		"""	Convert the model outputs into the different components of the
		loss function

		Args:
			output (tf.Tensor): The predicted values of the lensing parameters

		Returns:
			([tf.Tensor,...]): The list of tf.Tensor objects that correspond to
			the components of the loss function.
		"""
		# Start by dividing the output into the L_elements and the prediction
		# values.
		L_elements_len = int(self.num_params*(self.num_params+1)/2)
		y_pred, L_mat_elements = tf.split(output,
			num_or_size_splits=[self.num_params,L_elements_len],axis=-1)

		# Build the precision matrix and extract the diagonal part
		prec_mat, L_diag = self.construct_precision_matrix(L_mat_elements)

		# Return the mle, the precision matrix, and the diagonal values.
		return y_pred, prec_mat, L_diag

	def draw_samples(self,output,n_samps):
		"""	Draw samples from the predicted posterior defined by the networks
		outputs.

		Args:
			output (tf.Tensor): The predicted values of the lensing parameters
			n_samps (int): The number of samples to draw from each posterior.

		Returns:
			(np.array): An array with shape (n_samps,batch_size,n_params)
			containing draws from the posterior of each lens defined in
			the output.
		"""
		# The posterior is the mean with random noise.
		y_pred, prec_mat, _ = self.convert_output(output)
		y_pred = y_pred.numpy()
		prec_mat = prec_mat.numpy()
		cov_mat = np.linalg.inv(prec_mat)

		predict_samps = np.zeros((n_samps,)+y_pred.shape)

		for i in range(len(y_pred)):
			predict_samps[:,i,:] = np.random.multivariate_normal(y_pred[i],
				cov_mat[i],size=n_samps)
		return predict_samps

	def construct_precision_matrix(self,L_mat_elements):
		""" Take the matrix elements for the log cholesky decomposition and
		convert them to the precision matrix and return the value of
		the diagonal elements before exponentiation

		Args:
			L_mat_elements (tf.Tensor): A tensor of length
				num_params*(num_params+1)/2 that define the lower traingular
				matrix elements of the log cholesky decomposition

		Returns:
			((tf.Tensor,tf.Tensor)): Both the precision matrix and the diagonal
			elements (before exponentiation) of the log cholesky L matrix.
			Note that this second value is important for the posterior
			calculation.
		"""
		# First split the tensor into the elements that will populate each row
		cov_elements_split = tf.split(L_mat_elements,
			num_or_size_splits=self.split_list,axis=-1)
		# Before we stack these elements, we have to pad them with zeros
		# (corresponding to the 0s of the lower traingular matrix).
		cov_elements_stack = []
		pad_offset = 1
		for cov_element in cov_elements_split:
			# Use tf pad function since it's likely the fastest option.
			pad = tf.constant([[0,0],[0,self.num_params-pad_offset]])
			cov_elements_stack.append(tf.pad(cov_element,pad))
			pad_offset+=1
		# Stack the tensors to form our matrix. Use axis=-2 to avoid issues
		# with batches of matrices being passed in.
		L_mat = tf.stack(cov_elements_stack,axis=-2)
		# Pull out the diagonal part, and then (since we're using log
		# cholesky) exponentiate the diagonal.
		L_mat_diag = tf.linalg.diag_part(L_mat)
		L_mat = tf.linalg.set_diag(L_mat,tf.exp(L_mat_diag))
		# Calculate the actual precision matrix
		prec_mat = tf.matmul(L_mat,tf.transpose(L_mat,perm=[0,2,1]))

		return prec_mat, L_mat_diag

	@staticmethod
	def log_gauss_full(y_true,y_pred,prec_mat,L_diag):
		""" Return the negative log posterior of a Gaussian with full
		covariance matrix

		Args:
			y_true (tf.Tensor): The true values of the parameters
			y_pred (tf.Tensor): The predicted value of the parameters
			prec_mat: The precision matrix
			L_diag (tf.Tensor): The diagonal (non exponentiated) values of the
				log cholesky decomposition of the precision matrix

		Returns:
			(tf.Tensor): The TF graph for calculating the nlp

		Notes:
			This loss does not include the constant factor of 1/(2*pi)^(d/2).
		"""
		y_dif = y_true - y_pred
		return -tf.reduce_sum(L_diag,-1) + 0.5 * tf.reduce_sum(
			tf.multiply(y_dif,tf.reduce_sum(tf.multiply(tf.expand_dims(
				y_dif,-1),prec_mat),axis=-2)),-1)

	def loss(self,y_true,output):
		""" Returns the loss of the predicted parameters.

		Args:
			y_true (tf.Tensor): The true values of the parameters
			output (tf.Tensor): The predicted values of the lensing
				parameters. This should include 2*self.num_params parameters
				to account for the diagonal entries of our covariance matrix.
				Covariance matrix values are assumed to be in log space.

		Returns:
			(tf.Tensor): The loss function as a tf.Tensor.
		"""
		# Extract the outputs
		y_pred, prec_mat, L_diag = self.convert_output(output)

		# Add each possible flip to the loss list. We will then take the
		# minimum.
		loss_list = []
		for flip_mat in self.flip_mat_list:
			loss_list.append(self.log_gauss_full(y_true,
				tf.matmul(y_pred,flip_mat),prec_mat,L_diag))
		loss_stack = tf.stack(loss_list,axis=-1)
		return tf.reduce_min(loss_stack,axis=-1)



# -*- coding: utf-8 -*-
"""
Utility functions for simplifying interactions with lenstronomy classes.
"""
from lenstronomy.ImSim.Numerics.numerics_subframe import NumericsSubFrame


class PSFHelper():
	"""Class for simplifying interactions with the lenstronomy PSF class
	when using 2d images generated outside lenstronomy.

	Args:
		data_class (lenstronomy.Data.imaging_data.ImageData): An instance of
			the lenstronomy ImageData class, likely spawned from the
			lenstronomy DataAPI class. It will be used to define the grid
			onto which the PSF will be mapped.
		psf_model (lenstronomy.Data.psf.PSF): An instance of the lenstronomy
			PSF class that will be used to convolve the image.
		kwargs_numerics (dict): A dict containing the numerics kwargs to
			pass to lenstronomy
	"""

	def __init__(self,data_class,psf_model,kwargs_numerics):
		# Just initialize the NumericsSubframe object that we will later call
		# to perform the convolutions
		psf_model.set_pixel_size(data_class.pixel_width)
		self.image_numerics = NumericsSubFrame(pixel_grid=data_class,
			psf=psf_model,**kwargs_numerics)
		self.data_class = data_class

	def psf_model(self,image):
		"""Apply the psf model to the given image.

		Args:
			image (np.array): The image that will be convolved with the psf.
				It must have dimensions and resolution matching those specified
				in the production of the data_class provided when PSFHelper was
				initialized.

		Returns:
			(np.array): The image convolved with the psf.
		"""
		# Store the image shape for later
		im_shape = image.shape

		# Convolve the image in the 1d format required by the lenstronomy
		# class.
		conv_class = self.image_numerics.convolution_class

		if conv_class is None:
			# The psf type was NONE so no convolution should be done
			return image
		else:
			conv_flat = conv_class.convolution2d(image)
			# Reshape the image to be 2d if a 1d image was returned.
			return conv_flat.reshape(im_shape)
